# #############################################################################
# tests.py: Customized SCons tests
#
# Copyright 2006 the Ithildin Project.
# See the COPYING file for more information on licensing and use.
#
# $Id$
#
# #############################################################################

import os
import SCons.Errors

###############################################################################
# Feature tests

# Compiler type test: Check our compiler/version, bomb out for compilers
# that we know suck and add special flags for compilers where we need to.
def check_compiler_version(ctx):
    ctx.Message('Checking compiler/version...')
    if not ctx.env.has_key('CC'):
        ctx.Result('unknown')
        ctx.env.Exit(1)

    if ctx.env['CC'] == 'gcc':
        if not ctx.env.has_key('CCVERSION'):
            ctx.Result('unknown')
            ctx.env.Exit(1)
        ver = ctx.env['CCVERSION'].split('.')
        if ver[0] < 3:
            ctx.Result('gcc versions below 3.0 are not supported')
            ctx.env.Exit(1)

        # add -std=c99 .. gcc doesn't define __STDC_VERSION__ unless you add
        # a commandline flag asking it to use a specific standard (wtf?)
        ctx.env.PrependUnique(CCFLAGS = '-std=c99')

    if ctx.env.has_key('CCVERSION'):
        ctx.Result('%s (%s)' % (ctx.env['CC'], ctx.env['CCVERSION']))
    else:
        ctx.Result(ctx.env['CC'])

    return 1
# end of check_compiler_version

        
# compiler test: we require a C99 compiler.  It's (at least) 2006, time
# enough for people to get modern compilers.
def check_compiler_c_version(ctx):
    ctx.Message('Checking compiler for standard C version supported...')
    (res, data) = ctx.TryRun(extension = '.c', text = r'''
#include <stdio.h>
int main(void) {
#ifndef __STDC_VERSION__
    printf("unknown\n");
    return 0;
#else
    printf("%ld\n", __STDC_VERSION__);
    return 0;
#endif
}
''')

    data = data.strip()
    if res:
        ctx.Result(data)
        ver = int(data)
        if ver < 199901:
            print 'C99 support is required'
            ctx.env.Exit(1)
        return 1
    else:
        ctx.Result('none')
        print 'A standard C compiler is required'
        ctx.env.Exit(1)
# end of check_compiler_c_version

# Library test: OpenSSL.  Look for OpenSSL either in a predefined set of
# areas or explicitly in only one location.
openssl_prefixes = ['/usr/local', '/usr/ssl', '/usr/openssl',
    '/usr/local/ssl', '/usr/local/openssl' ]

def check_openssl(ctx, path = openssl_prefixes):
    ctx.Message('Looking for OpenSSL...')

    testcode = '''
#include <openssl/ssl.h>
int main(void) {
    SSL_accept(NULL);
}
'''

    if ctx.env.has_key('CPPPATH'): saveCPPPATH = ctx.env['CPPPATH']
    else: saveCPPPATH = []
    if ctx.env.has_key('LIBPATH'): saveLIBPATH = ctx.env['LIBPATH']
    else: saveLIBPATH = []
    if ctx.env.has_key('LIBS'): saveLIBS = ctx.env['LIBS']
    else: saveLIBS = []

    ctx.env.Append(LIBS = ['ssl', 'crypto'])

    # Try looking for OpenSSL in two ways:
    # If the directory isn't given, try just linking it using the standard
    # system-defined prefixes (this works almost all the time)
    # If it is given or if the above didn't work, then try walking the list
    # of prefixes to see if the requisite files are there
    if path == openssl_prefixes or path == [True]:
        if ctx.TryLink(extension = '.c', text = testcode):
            # got it, okay
            ctx.Result('found (system)')
            return 1
    
    # This will only work right on Unix platforms and it's probably not
    # really the right solution...unfortunately scons can't handle
    # removing stuff from env vars and I don't want to hack it, so here
    # we are...
    for p in path:
        if (os.path.isfile(p + '/include/openssl/ssl.h') and
            os.path.isfile(p + '/lib/libssl.a')):

            # This looks likely, test it out
            ctx.env.Append(CPPPATH = p + '/include')
            ctx.env.Append(LIBPATH = p + '/lib')

            if ctx.TryLink(extension = '.c', text = testcode):
                ctx.Result('found (%s)' % p)
                return 1
            else:
                # Failed for some reason, we presume this is a broken
                # OpenSSL install and let them know...
                print 'Found broken OpenSSL install in %s' % p
            
            # clean out any braindamage if we got here...
            ctx.env.Replace(CPPPATH = saveCPPPATH)
            ctx.env.Replace(LIBPATH = saveLIBPATH)

    # Obviously this has failed
    ctx.env.Replace(LIBS = saveLIBS)
    ctx.Result('not found')
    return 0
# end of check_openssl
        
def check_ipv6(ctx, functions = []):
    ctx.Message('Checking for IPv6 support...')
    if 'getaddrinfo' in functions:
        ctx.Result('found')
        return 1

    ctx.Result('not found')
    return 0
# end of check_ipv6

tests = {
    'CheckCCompiler': check_compiler_version,
    'CheckCVersion': check_compiler_c_version,
    'CheckIPv6': check_ipv6,
    'CheckOpenSSL': check_openssl,
}

###############################################################################
# Option validators
def PathIsDirCreateAccess(key, val, env):
    """Validate a path to ensure that it is a directory, if it does not
    exist try to create it unless we do not have access."""
    if os.path.isfile(val):
        m = 'Path for option %s is a file, not a directory: %s'
        raise SCons.Errors.UserError(m % (key, val))
    if not os.path.isdir(val):
        # check for access before os.makedirs()
        # This is kinda hard, we need to walk down the path they give us
        # until the provided directories cease to exist, then check the last
        # directory which exists for access (phew)
        dirname = os.path.dirname(val)
        while not os.path.exists(dirname):
            (dirname, rest) = os.path.split(dirname)
            if not rest: break # hit the root directory
        if os.path.isfile(dirname):
            m = 'Path for option %s includes a file %s'
            raise SCons.Errors.UserError(m % (key, val))
        if not os.access(dirname, os.R_OK | os.W_OK):
            m = 'No access to path for option %s (at %s)'
            raise SCons.Errors.UserError(m % (key, val))
        os.makedirs(val)
    elif not os.access(val, os.R_OK | os.W_OK):
        m = 'No access to path for option %s'
        raise SCons.Errors.UserError(m % (key))

def PathIsDirAccess(key, val, env):
    """Validate a path to ensure that it exists and is a directory.  If it
    does not exist or we have no access raise an exception."""
    if os.path.isfile(val):
        m = 'Path for option %s is a file, not a directory: %s'
        raise SCons.Errors.UserError(m % (key, val))
    if not os.path.isdir(val):
        m = "Path for option %s does not exist"
        raise SCons.Errors.UserError(m % (key))
    elif not os.access(val, os.R_OK | os.W_OK):
        m = 'No access to path for option %s'
        raise SCons.Errors.UserError(m % (key))

def PathCreateAccess(key, val, env):
    """Ensure that we can create a file or directory with the given name if
    it does not exist, or overwrite the existing one."""
    
    if os.path.exists(val) and not os.access(val, os.R_OK | os.W_OK):
        m = "Path for option %s is not writable"
        raise SCons.Errors.UserError(m % (key))
    elif not os.path.exists(val):
        dirname = os.path.dirname(val)
        while not os.path.exists(dirname):
            (dirname, rest) = os.path.split(dirname)
            if not rest: break # hit the root directory
        if not os.access(dirname, os.R_OK | os.W_OK):
            m = 'No access to path for option %s (at %s)'
            raise SCons.Errors.UserError(m % (key, val))

###############################################################################
# Utility functions
def AddConfInstall(env, dir, files):
    """Add an installer for a set of files.  We will copy the files with
    their original names if the destination files don't exist, otherwise
    we'll copy them in as .dist files."""

    for f in files:
        fname = os.path.basename(f)
        # always the '.dist' version
        env.InstallAs(dir + '/%s.dist' % fname, f)
        if not os.path.exists(dir + '/%s' % fname):
            env.Install(dir, f)

# vi:set ts=8 sts=4 sw=4 tw=76 et:
