/*
 * callback.h: Call-back functionality support
 * 
 * Copyright 2002 the Ithildin Project.
 * See the COPYING file for more information on licensing and use.
 * 
 * $Id$
 */

#ifndef CALLBACK_H
#define CALLBACK_H

/* All callback data is stored per-structure.  For a structure to be
 * callback-safe it must use the CALLBACK_SAFE macro in the structure
 * definition.  Then a callback against a structure may be initialized with
 * the below functions */

typedef bool (*callback_cancel_function)(void *);
#define CALLBACK_CANCEL_FUNCTION(func) bool func(void *data __UNUSED)

struct callback_entry {
    callback_cancel_function cancel;
    SLIST_ENTRY(callback_entry) lp;
};

#define CALLBACK_SAFE                                                         \
    unsigned int _callback_pending;                                           \
    SLIST_HEAD(, callback_entry) _callback_list;

/* The entire callback system ends up implemented in macros (just like
 * queue.h is) so that we can be "type safe".  Generally speaking when
 * callbacks are used the following should happen:
 * - A requestor should say "please call me back regarding this object using
 *   this function"
 * - The function taking the request should use callback_register
 * - The async code should execute and, if it calls back, should call
 *   callback_unregister just prior to doing this.
 * At any time during the above the cancellation function may be called via
 * cancel_callbacks.  In this case the cleanup of the object is handled by
 * cancel_callbacks, and the cancellation function should simply cease the
 * async operation related to the object.  cancel_callbacks should be called
 * by an object's release method.
 */
#define callback_register(obj, cancel) do {                                   \
    struct callback_entry *_cble = calloc(1, sizeof(struct callback_entry));  \
    SLIST_INSERT_HEAD(&obj->_callback_list, _callback_list_entry, lp);        \
    obj->_callback_pending++;                                                 \
} while (0)

#define callback_unregister(obj, cancel) do {                                 \
    struct callback_entry *_cble, *_cbletmp;                                  \
    assert(obj->_callback_pending > 0);                                       \
    SLIST_FOREACH_SAFE(_cble, &obj->_callback_list, lp, _cbletmp) {           \
        if (_cble->cancel == cancel) {                                        \
            obj->_callback_pending--;                                         \
            SLIST_REMOVE(&obj->_callback_list, _cble, struct callback_entry,  \
                         lp);                                                 \
            free(_cble);                                                      \
            break;                                                            \
        }                                                                     \
    }                                                                         \
    /* _cble will be NULL if and only if we traversed the loop and found no   \
       match.  We should not ever be called without a callback registered. */ \
    assert(_cble != NULL);                                                    \
} while (0)

/* Return the number of callbacks pending on an object */
#define callback_pending(obj) obj->_callback_pending

/* Cancel all callbacks for an object.  Should be called just before
 * invalidating said object. */
#define callback_cancel(obj) do {                                             \
    struct callback_entry *_cble;                                             \
    while (!SLIST_EMPTY(&obj->_callback_list)) {                              \
        _cble = SLIST_FIRST(&obj->_callback_list);                            \
        SLIST_REMOVE_HEAD(&obj->_callback_list, lp);                          \
        _cble->cancel(obj);                                                   \
        free(_cble);                                                          \
    }                                                                         \
    obj->_callback_pending = 0;                                               \
} while (0)

#endif
/* vi:set ts=8 sts=4 sw=4 tw=76 et: */
