/*
 * module.c: module system support functions
 * 
 * Copyright 2002-2006 the Ithildin Project.
 * See the COPYING file for more information on licensing and use.
 */

#include <ithildin/ithildin.h>

IDSTRING(rcsid, "$Id$");

#ifdef HAVE_DLFCN_H
# include <dlfcn.h>
#else
# ifdef HAVE_DL_H
#  include <dl.h>
# endif
#endif

/* create a module structure. */
static module_t *add_module_from_config(conf_entry_t *);
static module_t *create_module(char *);
static void destroy_module(module_t *);
static bool load_module_deps(module_t *, char **, int, bool);

/* this is called both at start-up and at any time when the conf is reloaded
 * to parse the 'modules' configuration section */
void build_module_list(void) {
    conf_list_t *list = conf_find_list("modules", me.confhead, 1);
    conf_entry_t *cep = NULL, *oldcep;
    module_t *m = NULL;
    module_t *last = NULL;

    if (list == NULL) {
        log_warn("there is no config entity for modules, this probably "
                "won't be a very functional daemon.");
        return;
    }

    /* new configuration stuff will be read in here if it is created */

    /* use the conf_find_next bits to find what we want here */
    oldcep = cep = conf_find("module", NULL, 0, list, 1);
    /* now pull in each load statement and parse it */
    while (cep != NULL) {
        m = add_module_from_config(cep);
        oldcep = cep = conf_find_next("module", NULL, 0, oldcep, list, 1);

        if (m == NULL)
            continue; /* no need to add it to the list.. */
        if (last == NULL)
            LIST_INSERT_HEAD(&me.modules, m, lp);
        else
            LIST_INSERT_AFTER(last, m, lp);
        last = m;
    }

    /* now load modules which requested autoloading.  we do this here instead
     * of above so we can read in all the module data before we load
     * anything. */ 
    LIST_FOREACH(m, &me.modules, lp) {
        if (m->flags & MODULE_FL_AUTOLOAD && !(m->flags & MODULE_FL_LOADED))
            load_module(m->name, m->flags);
    }
}

static module_t *add_module_from_config(conf_entry_t *cep) {
    conf_list_t *clp = NULL;
    module_t *m;
    bool new = false;
    char *s = NULL;
    char *dname = NULL;

    if (cep->type == CONF_TYPE_LIST) {
        /* check the 'string' value first, before checking the file
         * value. */
        clp = cep->list;
        if (cep->string != NULL)
            s = cep->string;
        else {
            s = conf_find_entry("file", clp, 1);
            if (s == NULL) {
                log_warn("module with no name configured");
                return NULL;
            }
        }
    } else
        s = cep->string;

    m = find_module(s);
    if (m == NULL) {
        m = create_module(s);
        new = true;
    }
    if (m == NULL)
        return NULL;

    if (clp != NULL) {
        if ((s = conf_find_entry("config", clp, 1)) != NULL) {
            /* Let the config code sort out the correct path.. */
            m->conffile = strdup(s);
        }

        if ((s = conf_find_entry("load", clp, 1)) != NULL) {
            if (str_conv_bool(s, false))
                m->flags |= MODULE_FL_AUTOLOAD;
        } else
            m->flags |= MODULE_FL_AUTOLOAD; /* default is to load. */
        
        /* allow module name to be overwritten (not sure why we'd want this,
         * but let's see if it helps.. */
        if (cep->string != NULL) {
            free(m->name);
            m->name = strdup(cep->string);
        }
    } else {
        /* We just got "module foo;"  I bet they want to autoload
         * it, so let's try that. */
        m->flags |= MODULE_FL_AUTOLOAD;

        /* let's also take a stab at guessing their module config file name.
         * We try two things: $confdir/modulename.conf and
         * $confdir/modulename/modulename.conf.  For modules with '.'s in
         * their name we replace the '.'s with path separators and try that
         * way, using the last portion of the module's name as the name of
         * the config file. */
        s = dname = strdup(m->name);
        while (*s != '\0') {
            if (*s == '.')
                *s = '/';
            s++;
        }

        s = malloc(PATH_MAX);
        /* first try the path without the additional directory layer */
        snprintf(s, PATH_MAX, "%s/%s.conf", me.conf_path, dname);
        if (access(s, R_OK) == 0)
            m->conffile = s;
        else {
            /* Now try the subdirectory method, this is a little weird since
             * we need to get just the last component of the module's name
             * (if it has dots in it) */
            snprintf(s, PATH_MAX, "%s/%s/%s.conf", me.conf_path, dname,
                    (strchr(m->name, '.') ? strrchr(m->name, '.') + 1 :
                     m->name));
            if (access(s, R_OK) == 0)
                m->conffile = s;
        }
        free(dname);
        if (m->conffile == NULL)
            free(s);
    }

    if (m->conffile != NULL) {
        if ((m->confdata = read_conf(m->conffile)) == NULL) {
            log_error("could not parse config for module %s, not loading it.",
                    m->name);
            destroy_module(m);
            return NULL;
        }
    }

    if (new)
        return m;

    return NULL;
}

/* this function unloads all modules rather indiscriminately.  it just starts
 * at the top of the load and starts dropping them. */
void unload_all_modules(void) {
    module_t *m;

    while ((m = LIST_FIRST(&me.modules)) != NULL) {
        unload_module(m->name);
        destroy_module(m);
    }
}

/* this function handles all the necessary work to load a module.  it opens the
 * dll, checks version headers, loads dependencies, then calls the module's
 * loader function (in that order). */
int load_module(char *name, int flags) {
    module_t *m = NULL;
    /* always ask for explicit symbol resolution.  every library a module
     * depends on should be *already linked against* the module in question.
     * we do not provide symbol help (because it is not reliably
     * cross-platform) */
    int rtld_flags = RTLD_NOW;
    char fn[33]; /* used for making function names */
    char *sn; /* short name */
    struct module_savedata *msdp;
    struct module_dependency *mdep;
    char **deplist = NULL;

    if ((m = find_module(name)) == NULL) {
        if (!(flags & MODULE_FL_CREATE)) {
            log_warn("attempt to load module %s not in modules list.", name);
            return 0;
        } else {
            module_t *last;
            m = create_module(name);
            if (m == NULL)
                return 0;
            last = LIST_FIRST(&me.modules);
            while (last != NULL) {
                if (LIST_NEXT(last, lp) == NULL)
                    break;
                last = LIST_NEXT(last, lp);
            }

            LIST_INSERT_AFTER(last, m, lp);
        }
    }

    if (!(m->flags & MODULE_FL_LOADED) && !(flags & MODULE_FL_QUIET))
        log_notice("loading module %s...", m->name);

    /* if we are being depend loaded make sure the module is set that way,
     * if we've already been depend-loaded once we don't bother to go
     * through the rest of the motions below. */
    if (flags & MODULE_FL_DEPENDLOAD) {
        if (m->flags & MODULE_FL_DEPENDLOAD && m->flags & MODULE_FL_LOADED)
            return 1; /* already depend loaded */
        m->flags |= MODULE_FL_DEPENDLOAD;
    }

    m->handle = dlopen(m->fullpath, rtld_flags);
    m->opencalls++;
    if (m->handle == NULL) {
        log_error("in %s: %s", m->name, dlerror());
        return 0;
    }

    if (m->flags & MODULE_FL_LOADED)
        return 1;
        
    /* ***********************************************************************
     * from this point on we are assured that m->opencalls is 1, that is,
     * that this is our first try at loading the module, so any failures can
     * safely be used to destroy the module/dlclose it/etc.
     * ***********************************************************************/

    m->header = (struct module_header *)dlsym(m->handle, "mheader");
    if (m->header == NULL) {
        log_error("in %s: %s", m->name, dlerror());
        dlclose(m->handle);
        m->opencalls = 0;
        return 0;
    }

    sn = strrchr(m->name, '.');
    if (sn == NULL)
        sn = m->name;
    else
        sn++;

    snprintf(fn, 32, "%s_loader", sn);
    m->load_function = (module_load_function)dlsym(m->handle, fn);

    /* Refuse to load modules compiled to use a module API version either
     * older or newer than the one we supply.  Also refuse to load modules
     * which are at the same version but have a newer revision.  Warn for
     * modules which have the same version but an older revision. */
    if (m->header->baseversion.version != MODULE_API_VERSION ||
            m->header->baseversion.revision > MODULE_API_REVISION) {
        log_error("module %s expects module API version %i.%i, we supply "
                "%i.%i.  Module will not be loaded.",
                name, m->header->baseversion.version,
                m->header->baseversion.revision,
                MODULE_API_VERSION, MODULE_API_REVISION);
        unload_module(m->name);
        return 0;
    } else if (m->header->baseversion.revision < MODULE_API_REVISION) {
        log_warn("module %s expects module API version %i.%i, we supply "
                "%i.%i.",
                name, m->header->baseversion.version,
                m->header->baseversion.revision,
                MODULE_API_VERSION, MODULE_API_REVISION);
    } /* ugh, okay, it's an alright version :) */

    /* load dependcies.  before we do, attest to the open-ness of this module.
     * this isn't, technically, true, since we haven't initialized it.
     * however, that can't be helped in the case of circular dependencies.
     * this is simply a tough area because they will try to load this module
     * and.. well.. we'll be right back here! :/ */
    m->flags |= MODULE_FL_LOADED | flags;

    if ((deplist = (char **)dlsym(m->handle, "mdepends")) != NULL) {
        if (!load_module_deps(m, deplist, flags, true)) {
            unload_module(m->name);
            return 0;
        }
    }
    if ((deplist = (char **)dlsym(m->handle, "msoftdepends")) != NULL) {
        if (!load_module_deps(m, deplist, flags, false))
            log_warn("some soft dependencies for %s failed to load, "
                     "continuing anyway.", m->name);
    }

    /* okay, hopefully this will stay here.  I've decided to hook the
     * 'load_module' event prior to load function being called so that other
     * subsystems can do any data setup necessary that the load function might
     * wish to rely on.  hope this works. ;)   this kind of sucks because it
     * might be totally false if the loader errors out.  I'm not sure how to
     * resolve this one, maybe we should hook the unload event right after?
     * Hrmm... */
    hook_event(me.events.load_module, m->name);

    if (m->load_function != NULL) {
        if (!m->load_function((m->flags & MODULE_FL_RELOADING ? 1 : 0),
                    &m->savedata, &m->confdata, m)) {
            unload_module(m->name);
            log_error("failed to load module %s", m->name);
            return 0;
        }
    }

    /* lookup and set the unload function late so that it cannot be used by
     * unload_module() when we call it above in failure cases. */
    snprintf(fn, 32, "%s_unloader", sn);
    m->unload_function = (module_unload_function)dlsym(m->handle, fn);

    /* go ahead and take care of destroying the saved data for them */
    while (!SLIST_EMPTY(&m->savedata)) {
        msdp = SLIST_FIRST(&m->savedata);
        SLIST_REMOVE_HEAD(&m->savedata, lp);
        free(msdp->data);
        free(msdp->name);
        free(msdp);
    }

    /* load "post" dependencies here. A "post" dependency is a module which
     * we should load after initializing the current module.  These are
     * mostly identicaly to soft dependencies (load failures don't hurt us)
     * and are just done at a different time. */
    if ((deplist = (char **)dlsym(m->handle, "mpostdepends")) != NULL) {
        if (!load_module_deps(m, deplist, flags, false))
            log_warn("some soft dependencies for %s failed to load, "
                     "continuing anyway.", m->name);
    }

    /* if we are reloading, walk the dependants list and load all those, as
     * well. */
    if (m->flags & MODULE_FL_RELOADING) {
        log_debug("reloaded module %s", m->name);
        m->flags &= ~MODULE_FL_RELOADING; /* reloaded.  (keep this above) */
        LIST_FOREACH(mdep, &m->deps, lp) {
            load_module(mdep->mod->name,
                    mdep->mod->flags | MODULE_FL_DEPENDLOAD | MODULE_FL_QUIET);
        }
    }

    return 1;
}

static bool load_module_deps(module_t *m, char **deplist, int flags,
        bool hardfail) {
    module_t *m2;
    bool allok = true;
    struct module_dependency *mdep;

    while (deplist != NULL && *deplist != NULL) {
        m2 = find_module(*deplist);
        if (m2 == NULL || !(m2->flags & MODULE_FL_DEPENDLOAD)) {
            if (m2 == NULL)
                log_notice("loading dependant module %s for %s", *deplist,
                        m->name);
            if (!load_module(*deplist, flags | MODULE_FL_DEPENDLOAD |
                    MODULE_FL_CREATE | MODULE_FL_QUIET)) {
                if (hardfail == true)
                    return false;
                else
                    allok = false;
            }
            m2 = find_module(*deplist);
        }
        deplist++; /* move on to next module */

        if (hardfail) {
            /* m2 is guaranteed in this case */
            mdep = malloc(sizeof(struct module_dependency));
            mdep->mod = m;
            LIST_INSERT_HEAD(&m2->deps, mdep, lp);
        } else
            continue; /* do not link soft dependencies */
    }

    return allok;
}

module_t *find_module(char *name) {
    module_t *m;

    LIST_FOREACH(m, &me.modules, lp) {
        if (!strcmp(m->name, name))
            return m;
    }

    return NULL;
}

void *lookup_module_symbol(char *name, char *sym) {
    module_t *mod;
    if (*name == '\0')
        return module_symbol(NULL, sym);
    else {
        mod = find_module(name);
        if (mod != NULL)
            return module_symbol(mod, sym);
    }

    return NULL;
}

void *module_symbol(module_t *mod, const char *sym) {
    return dlsym(mod->handle, (char *) sym);
}

int unload_module(char *name) {
    module_t *m = find_module(name);
    module_t *m2;
    struct module_dependency *mdep, *mdep2;

    if (m == NULL) {
        log_warn("attempt to unloaded nonexistant module %s", name);
        return 0;
    }

    /* sure, we unloaded it!  since this may be done automatically to clean
     * up stale dependencies, and may happen more than once (possible,
     * shrug), we just silently ignore this */
    if (!(m->flags & MODULE_FL_LOADED))
        return 1; 

    /* tag the module as unloaded so we don't try and unload it again.  at this
     * point everything below ought to be successful, soooo. */
    m->flags &= ~(MODULE_FL_LOADED|MODULE_FL_DEPENDLOAD);

    /* go through their list of dependencies and unload them.  if we are
     * actually reloading, set the reload flag on all of them prior to the
     * unload call. */
    mdep = LIST_FIRST(&m->deps);
    while (mdep != NULL) {
        /* we have different behaviors depending on what we're doing.  if this
         * is a reload, we know our entries won't disappear, if it isn't, we
         * know they will! */
        if (m->flags & MODULE_FL_RELOADING) {
            mdep2 = LIST_NEXT(mdep, lp);
            if (mdep->mod->flags & MODULE_FL_LOADED) {
                mdep->mod->flags |= MODULE_FL_RELOADING;
                unload_module(mdep->mod->name);
            }
            mdep = mdep2;
        } else {
            if (mdep->mod->flags & MODULE_FL_LOADED)
                unload_module(mdep->mod->name);
            else {
                log_warn("unload_module(%s): unloaded dependency for %s",
                        m->name, mdep->mod->name);
                LIST_REMOVE(mdep, lp);
                free(mdep); /* um..? */
            }
            mdep = LIST_FIRST(&m->deps);
        }
    }

    /* similar to the way it works in load_module(), hook here, before we call
     * the unload function, so that everything is cleaned up prior to the
     * module really disappearing. */
    hook_event(me.events.unload_module, m->name);

    /* clear sailing from here, close the handle and clear out the data */
    if (m->unload_function != NULL)
        m->unload_function((m->flags & MODULE_FL_RELOADING ? 1 : 0),
                &m->savedata);
        
    while (m->opencalls) {
#ifdef DEBUG
        /* dmalloc doesn't like it when we close libraries while it's shutting
         * down.  okay, well, if me.shutdown is true, don't dlclose, but do
         * everything else.  Valgrind dislikes this too, let's just disable
         * it in the debug case. */
        if (!me.shutdown)
#endif
        if (dlclose(m->handle))
            log_error("dlclose(%s): %s", m->name, dlerror());
        m->opencalls--;
    }

    /* now for each of our modules, if they have a dependency for this module,
     * remove it from their listing. */
    if (!(m->flags & MODULE_FL_RELOADING)) {
        LIST_FOREACH(m2, &me.modules, lp) {
            mdep = LIST_FIRST(&m2->deps);
            while (mdep != NULL) {
                mdep2 = LIST_NEXT(mdep, lp);
                if (mdep->mod == m) {
                    LIST_REMOVE(mdep, lp);
                    free(mdep);
                }
                mdep = mdep2;
            }
        }
    }

    log_debug("unloaded module %s", m->name);

    return 1;
}

int reload_module(char *name) {
    module_t *m = find_module(name);

    if (m == NULL)
        return 0;
    m->flags |= MODULE_FL_RELOADING;
    me.reloads++;

    return 1; /* the module will get reloaded later. */
}

void do_module_reloads(void) {
    module_t *m;

    LIST_FOREACH(m, &me.modules, lp) {
        if (!(m->flags & MODULE_FL_RELOADING))
            continue; /* skip */

        if (m->flags & MODULE_FL_LOADED)
            if (unload_module(m->name) == 0)
                return;

        load_module(m->name, m->flags);
    }
}

int module_loaded(char *name) {
    module_t *m = find_module(name);

    if (m != NULL && m->flags & MODULE_FL_LOADED)
        return 1;

    return 0;
}

/* this function adds data of the given size with the given name into the given
 * list.  it is useful to call this when a module is being unloaded to save its
 * data. */
void add_module_savedata(struct module_savedata_list *list, const char *name,
        size_t size, const void *data) {
    struct module_savedata *amsdp = malloc(sizeof(struct module_savedata));

    amsdp->name = strdup(name);
    amsdp->size = size;
    amsdp->data = malloc(amsdp->size);
    memcpy(amsdp->data, data, amsdp->size);
    SLIST_INSERT_HEAD(list, amsdp, lp);               
}

/* this function attempts to copy the data from a given module savedata list
 * with the given name into the given address.  If it finds the data, it copies
 * it in and returns the number of bytes copied, otherwise it returns 0. */
size_t get_module_savedata(struct module_savedata_list *list, const char *name,
        void *data) {
    struct module_savedata *gmsdp;
    SLIST_FOREACH(gmsdp, list, lp) {
        if (!strcmp(gmsdp->name, name)) {
            memcpy(data, gmsdp->data, gmsdp->size);
            return gmsdp->size;
        }
    }                       

    return 0;
}

static module_t *create_module(char *name) {
    module_t *m;
    char *s;

    m = calloc(1, sizeof(module_t));
    m->fullpath = malloc(PATH_MAX);
    m->name = strdup(name);

    /* strip off the prefix and suffix, if they exist. */
    if (SHLIBPREFIX[0]) {
        size_t pfxlen = strlen(SHLIBPREFIX);
        size_t namelen = strlen(m->name);
        if (!strncasecmp(m->name, SHLIBPREFIX, pfxlen)) {
            memmove(m->name, m->name + pfxlen, namelen - pfxlen);
            m->name[namelen - pfxlen] = '\0';
        }
    }
    if (SHLIBSUFFIX[0]) {
        s = strrchr(m->name, SHLIBSUFFIX[0]);
        if (s != NULL) {
            if (!strcmp(s, SHLIBSUFFIX))
                *s = '\0';
        }
    }

    sprintf(m->fullpath, "%s/%s%s%s", me.lib_path, SHLIBPREFIX, m->name,
            SHLIBSUFFIX);
    if (access(m->fullpath, R_OK | X_OK) == -1) {
        log_error("unable to open module %s from %s: %s", m->name,
                m->fullpath, strerror(errno));
        destroy_module(m);
        return NULL;
    }
    return m;
}

static void destroy_module(module_t *mod) {

    /* only remove it if it's in the list */
    if (find_module(mod->name) != NULL)
        LIST_REMOVE(mod, lp);
    free(mod->name);
    free(mod->fullpath);
    if (mod->conffile != NULL)
        free(mod->conffile);
    if (mod->confdata != NULL)
        destroy_conf_branch(mod->confdata);
    free(mod);
}

/* below here live the declarations for 'module data extensions' (mdext).  the
 * gist of this is that you create an mdext header for your structure type, and
 * then in each structure place a char pointer called 'mdext'.  now when a
 * module wants to register extra data in another module's structure, it can
 * ask for an mdext offset, with a given size, and can then use that mdext
 * offset to access into the structure.  modules are responsible for properly
 * allocating mdext space, and should provide hooks for expanding and shrinking
 * mdext data. */

#define mdext_realloc(header, data) do {                                      \
    if (header->size != 0)                                                    \
        data = realloc(data, header->size);                                   \
    else {                                                                    \
        if (data != NULL)                                                     \
            free(data);                                                       \
        data = NULL;                                                          \
    }                                                                         \
} while (0)

/* this creates an mdext header.  you should have one header per structure
 * type, not per structure.  this is used as the head of the tracking data for
 * that structure. */
struct mdext_header *create_mdext_header(mdext_iter_function iter) {
    struct mdext_header *mdhp = NULL;

    mdhp = calloc(1, sizeof(struct mdext_header));
    mdhp->create = create_event(EVENT_FL_NORETURN);
    mdhp->destroy = create_event(EVENT_FL_NORETURN);
    mdhp->iter = iter;

    return mdhp;
}

/* this destroys an mdext header and all associated data. */
void destroy_mdext_header(struct mdext_header *header) {
    struct mdext_item *mip, *mip2;

    /* wipe out any remaining items. */
    mip = SLIST_FIRST(&header->items);
    while (mip != NULL) {
        mip2 = SLIST_NEXT(mip, lp);
        destroy_mdext_item(header, mip);
        mip = mip2;
    }
    destroy_event(header->create);
    destroy_event(header->destroy);
    free(header);
}

/* this creates a new mdext item, and allocates new space for it in all the
 * structures handled with the iteration function in the header. */
struct mdext_item *create_mdext_item(struct mdext_header *header,
        size_t size) {
    struct mdext_item *mip;
    char **data;
    char *rhold = NULL; /* data held for iter */

    mip = calloc(1, sizeof(struct mdext_item));
    
    mip->size = size;
    mip->offset = header->size;
    header->size += mip->size;

    /* go through and resize everything.  also, zero-fill our new section. */
    while ((data = header->iter(&rhold)) != NULL) {
        mdext_realloc(header, *data);
        memset(*data + mip->offset, 0, mip->size);
    }

    SLIST_INSERT_HEAD(&header->items, mip, lp);
    return mip;
}

/* this function destroys an mdext_item, and also handles resizing and
 * re-aligning the offsets in each structure and item. */
void destroy_mdext_item(struct mdext_header *header, struct mdext_item *item) {
    struct mdext_item *mip;
    struct mdext_item *old;
    char **data;
    char *rhold = NULL; /* data held for iter */

    SLIST_FOREACH(mip, &header->items, lp) {
        if (mip == item)
            break;
    }
    if (mip == NULL) {
        log_warn("tried to free an mdext_item from an mdext_header it didn't "
                "belong to!");
        return;
    }

    /* this is where it gets tricky.  we have to shift memory down, then
     * reallocate to get rid of the extra space no longer being occupied.  the
     * way we do this is by doing a memmove at the offset for the rest of the
     * data, then calling realloc.  of course, we also have to slide down the
     * offsets for everything with an offset greater than the offset of the
     * item we're deleting, too. */

    /* if we need to move (that is, if offset+size isn't equivalent to the
     * total size (making this the last item)) then do so. */
    if (mip->offset + mip->size != header->size) {
        while ((data = header->iter(&rhold)) != NULL) 
            memmove(*data + mip->offset, *data + (mip->offset + mip->size),
                    (header->size - mip->offset - mip->size));
    }
    header->size -= mip->size;
    rhold = NULL;
    while ((data = header->iter(&rhold)) != NULL)
        mdext_realloc(header, *data);

    old = mip;
    SLIST_REMOVE(&header->items, old, mdext_item, lp);
    SLIST_FOREACH(mip, &header->items, lp) {
        if (mip->offset > old->offset)
            mip->offset -= old->size; /* move down. */
    }
    free(old);
}

/* this function allocates the data for an object with the given mdext header.
 * it also calls the "create" hook, so don't call mdext_alloc more than once
 * unless you want trouble. ;) */
char *mdext_alloc(struct mdext_header *header) {
    char *data = NULL;

    if (header->size) { 
        data = calloc(1, header->size);
        hook_event(header->create, data);
    }

    return data;;
}

/* this function frees the data for ab object with the given mdext header.  it
 * also calls the "destroy" hook. */
void mdext_free(struct mdext_header *header, char *data) {

    if (data != NULL) {
        hook_event(header->destroy, data);
        free(data);
    }
}
/* vi:set ts=8 sts=4 sw=4 tw=76 et: */
