/*
 * chanmode.c: channel mode management code
 * 
 * Copyright 2002-2007 the Ithildin Project.
 * See the COPYING file for more information on licensing and use.
 * 
 * This file provides both the basic functions necessary to manage channel
 * modes and some of the intrinsic channel modes for IRC (mostly as defined
 * in rfc1459).  Additional modes can be created through other modules.
 */

#include <ithildin/ithildin.h>

#include "ircd.h"

IDSTRING(rcsid, "$Id$");

/* lots of static functions here */
static void chanmode_build_lists(void);
static int check_bans(const struct channel_ban_list *, const char *,
        const char *, const char *, const char *, const char *);
static CHANMODE_FUNC(chanmode_ban);
static CHANMODE_QUERY_FUNC(chanmode_ban_query);
static HOOK_FUNCTION(can_join_mode_b);
static HOOK_FUNCTION(can_send_mode_b);
static HOOK_FUNCTION(can_nick_mode_b);

static HOOK_FUNCTION(can_send_mode_m);
static HOOK_FUNCTION(can_send_mode_n);
static HOOK_FUNCTION(can_show_mode_s);

static HOOK_FUNCTION(can_act_mode_ov);

static CHANMODE_FUNC(chanmode_key);
static CHANMODE_QUERY_FUNC(chanmode_key_query);
static HOOK_FUNCTION(can_join_mode_k);

static CHANMODE_FUNC(chanmode_limit);
static CHANMODE_QUERY_FUNC(chanmode_limit_query);
static HOOK_FUNCTION(can_join_mode_l);

/* This function either initializes the chanmode system or handles updates
 * in the case of a reload. */
void chanmode_init(bool reload) {
    uint64_t ui64 = 1;
    unsigned char c;

#define MODELOOP(from, to) do {                                               \
    for (c = from;c <= to;c++, ui64 <<= 1) {                                  \
        ircd.cmodes.modes[c].mode = c;                                        \
        ircd.cmodes.modes[c].mask = ui64;                                     \
        ircd.cmodes.modes[c].avail = 1;                                       \
    }                                                                         \
} while (0)

    if (reload == false) {
        MODELOOP('a', 'z'); /* set a-z modes */
        MODELOOP('A', 'Z'); /* and A-Z modes */
        MODELOOP('0', '9'); /* and 0-9 modes */

        /* now allocate our various channel modes */
        ui64 = 0;
        ircd.cmodes.mode_ban = chanmode_request('b', &c, CHANMODE_FL_A,
                chanmode_ban, chanmode_ban_query,
                sizeof(struct channel_ban_list), NULL);
        ircd.cmodes.mode_key = chanmode_request('k', &c, CHANMODE_FL_B,
                chanmode_key, chanmode_key_query, PASSWDLEN + 1, NULL);
        ircd.cmodes.mode_limit = chanmode_request('l', &c, CHANMODE_FL_C,
                chanmode_limit, chanmode_limit_query, sizeof(uint32_t), NULL);
        ircd.cmodes.mode_mod = chanmode_request('m', &c, CHANMODE_FL_D,
                chanmode_flag, chanmode_flag_query, 0, NULL);
        ircd.cmodes.mode_nextern = chanmode_request('n', &c, CHANMODE_FL_D,
                chanmode_flag, chanmode_flag_query, 0, NULL);
        ircd.cmodes.mode_op = chanmode_request('o', &c, CHANMODE_FL_PREFIX,
                chanmode_uflag, chanmode_uflag_query, 0, "@");
        ircd.cmodes.mode_secret = chanmode_request('s', &c, CHANMODE_FL_D,
                chanmode_flag, chanmode_flag_query, 0, NULL);
        ircd.cmodes.mode_voice = chanmode_request('v', &c, CHANMODE_FL_PREFIX,
                chanmode_uflag, chanmode_uflag_query, 0, "+");
    } else {
        chanmode_update_funcs(ircd.cmodes.mode_ban, chanmode_ban,
                chanmode_ban_query);
        chanmode_update_funcs(ircd.cmodes.mode_key, chanmode_key,
                chanmode_key_query);
        chanmode_update_funcs(ircd.cmodes.mode_limit, chanmode_limit,
                chanmode_limit_query);
        chanmode_update_funcs(ircd.cmodes.mode_mod, chanmode_flag,
                chanmode_flag_query);
        chanmode_update_funcs(ircd.cmodes.mode_nextern, chanmode_flag,
                chanmode_flag_query);
        chanmode_update_funcs(ircd.cmodes.mode_op, chanmode_uflag,
                chanmode_uflag_query);
        chanmode_update_funcs(ircd.cmodes.mode_mod, chanmode_flag,
                chanmode_flag_query);
        chanmode_update_funcs(ircd.cmodes.mode_voice, chanmode_uflag,
                chanmode_uflag_query);
    }

    add_hook(ircd.events.can_join_channel, can_join_mode_b);
    add_hook(ircd.events.can_send_channel, can_send_mode_b);
    add_hook(ircd.events.can_nick_channel, can_nick_mode_b);

    add_hook(ircd.events.can_send_channel, can_send_mode_m);

    add_hook(ircd.events.can_send_channel, can_send_mode_n);

    add_hook(ircd.events.can_see_channel, can_show_mode_s);

    add_hook(ircd.events.can_join_channel, can_join_mode_k);
    add_hook(ircd.events.can_join_channel, can_join_mode_l);

    add_hook(ircd.events.can_send_channel, can_act_mode_ov);
    add_hook(ircd.events.can_nick_channel, can_act_mode_ov);

    ui64 = MAX_BANS_PER_CHANNEL;
    add_isupport("MAXBANS", ISUPPORT_FL_INT, (char *)&ui64);
}

void chanmode_deinit(bool reload) {

    if (reload == false) {
        chanmode_release(ircd.cmodes.mode_ban);
        chanmode_release(ircd.cmodes.mode_key);
        chanmode_release(ircd.cmodes.mode_limit);
        chanmode_release(ircd.cmodes.mode_mod);
        chanmode_release(ircd.cmodes.mode_nextern);
        chanmode_release(ircd.cmodes.mode_op);
        chanmode_release(ircd.cmodes.mode_secret);
        chanmode_release(ircd.cmodes.mode_voice);
    }

    remove_hook(ircd.events.can_join_channel, can_join_mode_b);
    remove_hook(ircd.events.can_send_channel, can_send_mode_b);
    remove_hook(ircd.events.can_nick_channel, can_nick_mode_b);

    remove_hook(ircd.events.can_send_channel, can_send_mode_m);

    remove_hook(ircd.events.can_send_channel, can_send_mode_n);

    remove_hook(ircd.events.can_see_channel, can_show_mode_s);

    remove_hook(ircd.events.can_join_channel, can_join_mode_k);
    remove_hook(ircd.events.can_join_channel, can_join_mode_l);

    remove_hook(ircd.events.can_send_channel, can_act_mode_ov);
    remove_hook(ircd.events.can_nick_channel, can_act_mode_ov);

    del_isupport(find_isupport("MAXBANS"));
}

uint64_t chanmode_request(unsigned char suggested, unsigned char *actual,
        int flags, chanmode_func changefn, chanmode_query_func queryfn,
        size_t extra, void *extdata) {
    struct chanmode *md = NULL;
    int i, j;
    unsigned char c;

    /* try and find the first/best available mode */
    if (ircd.cmodes.modes[suggested].avail)
        md = &ircd.cmodes.modes[suggested];
    else {
        if (islower(suggested))
            c = toupper(suggested);
        else
            c = tolower(suggested);

        if (ircd.cmodes.modes[c].avail)
            md = &ircd.cmodes.modes[c];
        else {
            for (i = 0;i < 256;i++) {
                if (ircd.cmodes.modes[i].avail && ircd.umodes.modes[i].mask) {
                    md = &ircd.cmodes.modes[i];
                    break;
                }
            }

            if (md == NULL)
                return 0;
        }
    }

    /* fill in the stuff */
    md->avail = 0;
    md->umask = 0;
    md->prefix = '\0';
    md->changefunc = changefn;
    md->queryfunc = queryfn;
    md->flags = flags;
    md->mdi = NULL;

    if (flags == CHANMODE_FL_PREFIX) {
        /* we have to find the first free flag, if any, to use.  to do
         * this, for each value from 0x1 to 0x80 moving in powers of two,
         * we have to check if any of the channel modes are using it.  fun
         * stuff. */
        for (i = 0;i < 16;i++) {
            for (j = 0;j < 256;j++) {
                if (ircd.cmodes.modes[j].umask == 1 << i)
                    break;
            }
            if (j == 256)
                break; /* we have a weiner */
        }
        if (i == 16) {
            /* couldn't get a mode.  yuck. */
            md->avail = 1;
            return 0;
        }
        md->umask = 1 << i;
        md->prefix = *(unsigned char *)extdata;
        ircd.cmodes.pfxmap[md->prefix] = md; /* map back */
    }

    if (extra > 0)
        md->mdi = create_mdext_item(ircd.mdext.channel, extra);
    
    chanmode_build_lists();

    *actual = md->mode;
    return md->mask;
}

/* this function very simply releases a mode and calls the list rebuilder. */
void chanmode_release(unsigned char mode) {
    channel_t *chan;
    int dummy;
    
    /* clear the mode from all the channels .. */
    LIST_FOREACH(chan, ircd.lists.channels, lp)
        ircd.cmodes.modes[mode].changefunc(NULL, chan, mode, CHANMODE_CLEAR, NULL, &dummy);

    /* we don't touch the mode or mask members because those are set elsewhere
     * and need to be kept the way they are. */
    ircd.cmodes.modes[mode].avail = 1; /* mark it as available */

    ircd.cmodes.modes[mode].umask = 0;
    if (ircd.cmodes.modes[mode].prefix != '\0') {
        ircd.cmodes.pfxmap[ircd.cmodes.modes[mode].prefix] = NULL;
        ircd.cmodes.modes[mode].prefix = '\0';
    }
    ircd.cmodes.modes[mode].changefunc = NULL;
    ircd.cmodes.modes[mode].queryfunc = NULL;
    ircd.cmodes.modes[mode].flags = 0;
    if (ircd.cmodes.modes[mode].mdi != NULL)
        destroy_mdext_item(ircd.mdext.channel, ircd.cmodes.modes[mode].mdi);
    chanmode_build_lists();
}

void chanmode_update_funcs(unsigned char mode, chanmode_func changefunc,
        chanmode_query_func queryfunc) {
    struct chanmode *cmp = &ircd.cmodes.modes[mode];

    if (cmp->avail || !cmp->mask) /* only if it exists.. */
        return;

    cmp->changefunc = changefunc;
    cmp->queryfunc = queryfunc;
}

static void chanmode_build_lists(void) {
    int i;
    unsigned char imodes[256];
    unsigned char *s;

    /* fill in the 'avail' modestring thing. */
    s = ircd.cmodes.avail;
    for (i = 0;i < 256;i++) {
        if (!ircd.cmodes.modes[i].avail && ircd.cmodes.modes[i].mask)
            *s++ = ircd.cmodes.modes[i].mode;
    }
    *s = '\0';

    /* now do the prefix thing.  we make two passes to make it easier. */
    s = ircd.cmodes.pmodes;
    for (i = 0;i < 256;i++) {
        if (!ircd.cmodes.modes[i].avail && ircd.cmodes.modes[i].prefix)
            *s++ = ircd.cmodes.modes[i].mode;
    }
    *s = '\0';

    s = ircd.cmodes.prefix;
    *s++ = '(';
    for (i = 0;i < 256;i++) {
        if (!ircd.cmodes.modes[i].avail && ircd.cmodes.modes[i].prefix)
            *s++ = ircd.cmodes.modes[i].mode;
    }
    *s++ = ')';
    for (i = 0;i < 256;i++) {
        if (!ircd.cmodes.modes[i].avail && ircd.cmodes.modes[i].prefix)
            *s++ = ircd.cmodes.modes[i].prefix;
    }
    *s = '\0';

    /* now do a/b/c/d/e */
    s = imodes;
#define CHANMODE_BUILD_TYPE(_flg) do {                                        \
    for (i = 0;i < 256;i++) {                                                 \
        if (!ircd.cmodes.modes[i].avail &&                                    \
                ircd.cmodes.modes[i].flags & _flg)                            \
            *s++ = ircd.cmodes.modes[i].mode;                                 \
    }                                                                         \
    *s++ = ',';                                                               \
} while (0)

    CHANMODE_BUILD_TYPE(CHANMODE_FL_A);
    CHANMODE_BUILD_TYPE(CHANMODE_FL_B);
    CHANMODE_BUILD_TYPE(CHANMODE_FL_C);
    CHANMODE_BUILD_TYPE(CHANMODE_FL_D);
    CHANMODE_BUILD_TYPE(CHANMODE_FL_E);
    *--s = '\0'; /* wipe out that last comma */
    /* also, if s - 1 is a comma (meaning we have no 'E' modes, delete that
     * comma since 'E' is non-standard */
    if (*(s - 1) == ',')
        *--s = '\0';

    /* add these two in for special mode support stuff */
    add_isupport("PREFIX", ISUPPORT_FL_STR, (char *)ircd.cmodes.prefix);
    add_isupport("CHANMODES", ISUPPORT_FL_STR, (char *)imodes);
}

int chanmode_set(unsigned char mode, client_t *cli, channel_t *chan, char *arg,
        int *argused) {
    struct chanmode *cmp = &ircd.cmodes.modes[mode];

    if (cmp->avail || !cmp->mask) /* only if it exists.. */
        return CHANMODE_NONEX;

    /* otherwise, just try and set the mode with the given functions, returning
     * the value as requested */
    return cmp->changefunc(cli, chan, mode, CHANMODE_SET, arg, argused);
}

int chanmode_setprefix(unsigned char prefix, channel_t *chan, char *arg,
        int *argused) {
    struct chanmode *cmp = ircd.cmodes.pfxmap[prefix];

    if (cmp == NULL)
        return CHANMODE_NONEX;

    /* found it, now set it */
    return cmp->changefunc(NULL, chan, cmp->mode, CHANMODE_SET, arg, argused);
}

int chanmode_unset(unsigned char mode, client_t *cli, channel_t *chan,
        char *arg, int *argused) {
    struct chanmode *cmp = &ircd.cmodes.modes[mode];

    if (cmp->avail || !cmp->mask) /* only if it exists.. */
        return CHANMODE_NONEX;

    /* otherwise, just try and set the mode with the given functions, returning
     * the value as requested */
    return cmp->changefunc(cli, chan, mode, CHANMODE_UNSET, arg, argused);
}

int chanmode_unsetprefix(unsigned char prefix, channel_t *chan, char *arg,
        int *argused) {
    struct chanmode *cmp = ircd.cmodes.pfxmap[prefix];

    if (cmp == NULL)
        return CHANMODE_NONEX;

    /* found it, now unset it */
    return cmp->changefunc(NULL, chan, cmp->mode, CHANMODE_UNSET, arg, argused);
}

int chanmode_query(unsigned char mode, channel_t *chan, char *arg,
        int *argused, void **state) {
    struct chanmode *cmp = &ircd.cmodes.modes[mode];

    if (cmp->avail || !cmp->mask) /* only if it exists.. */
        return CHANMODE_NONEX;

    /* let's ask 'em about it. */
    return cmp->queryfunc(chan, mode, arg, argused, state);
}

int chanmode_isprefix(channel_t *chan, client_t *cli, unsigned char prefix) {
    struct chanlink *clp;
    struct chanmode *cmp = ircd.cmodes.pfxmap[prefix];

    if (cmp == NULL)
        return 0; /* no such mode. */

    /* found it, now see if it's set */
    LIST_FOREACH(clp, &chan->users, lpchan) {
        if (clp->cli == cli)
            return chanlink_ismode(clp, cmp->mode);
    }

    /* not found, no status */
    return 0;
}

char *chanmode_getprefixes(channel_t *chan, client_t *cli) {
    static unsigned char pfx[64];
    struct chanlink *clp = find_chan_link(cli, chan);
    unsigned char *s = ircd.cmodes.pmodes;
    int i = 0;

    if (clp != NULL) {
        while (*s != '\0') {
            if (clp->flags & ircd.cmodes.modes[*s].umask)
                pfx[i++] = ircd.cmodes.modes[*s].prefix;
            s++;
        }
    }

    pfx[i] = '\0';
    return (char *)pfx;
}

/* this function returns two strings.  the first is all the modes set on the
 * channel.  the second is any arguments for the modes, but may or may not be
 * truncated. */
const char **chanmode_getmodes(channel_t *chan) {
    static const char *ret[2];
    static unsigned char modes[64];
    static char *modebuf = NULL;
    static size_t mbsize = 64;
    int optused, len = 0, m = 1;
    unsigned char *i = ircd.cmodes.avail;
    void *state;

    if (modebuf == NULL)
        modebuf = malloc(mbsize);

    modes[0] = '+';
    *modebuf = '\0';

    while (*i != '\0') {
        state = NULL;

        if (chan->modes & ircd.cmodes.modes[*i].mask) {
            optused = mbsize - len;
            while (!chanmode_query(*i, chan, modebuf + len, &optused, &state)) {
                if (optused < 0) {
                    mbsize += -optused + 2;
                    modebuf = realloc(modebuf, mbsize);
                    optused = mbsize - len;
                    continue;
                }
                modes[m++] = *i;
                if (optused > 0) {
                    len += optused;
                    modebuf[len++] = ' ';
                    modebuf[len] = '\0';
                }
            }
        }
        i++;
    }
    modes[m] = '\0';
    if (len > 0)
        modebuf[--len] = '\0'; /* nerf that extra space */

    ret[0] = (char *)modes;
    ret[1] = modebuf;
    return ret;
}

/* Below are four convenience functions for modes which fit the generic
 * category of either channel 'flags' (on/off or type 'D') or channel user
 * 'prefixes' (uflags like +o/+v).  If your flags don't need special
 * behavior to be set/unset use these. */
CHANMODE_FUNC(chanmode_flag) {
    
    *argused = 0;
    if (cli != NULL && !CHANOP(cli, chan) && !CLIENT_MASTER(cli))
        return ERR_CHANOPRIVSNEEDED;

    switch (set) {
        case CHANMODE_SET:
            chanmode_setflag(chan, mode);
            break;
        case CHANMODE_UNSET:
            chanmode_unsetflag(chan, mode);
            break;
    }
    return CHANMODE_OK;
}

CHANMODE_QUERY_FUNC(chanmode_flag_query) {
    *argused = 0;

    if (*state == NULL)
        *state = (void *)0x1; /* dummy pointer to nowhere */
    else
        return CHANMODE_FAIL; /* they already asked */

    if (!chanmode_isset(chan, mode))
        return CHANMODE_FAIL; /* not set */
    return CHANMODE_OK; /* it's set */
}

CHANMODE_FUNC(chanmode_uflag) {
    struct chanlink *clp;
    client_t *cp;

    if (cli != NULL && !CHANOP(cli, chan) && !CLIENT_MASTER(cli))
        return ERR_CHANOPRIVSNEEDED;

    /* error check stuff first */
    if (set != CHANMODE_CLEAR) {
        if (arg == NULL)
            return CHANMODE_NOARG;
        *argused = 1; /* otherwise we use the argument */
    } else
        return CHANMODE_OK; /* we don't have anything to clear */

    /* chase for +o/+v to avoid desynchs */
    cp = client_get_history(arg, 0);
    if (cp == NULL) {
        if (cli != NULL)
            sendto_one(cli, RPL_FMT(cli, ERR_NOSUCHNICK), arg);
        return CHANMODE_FAIL;
    }
    if (arg != cp->nick)
        strcpy(arg, cp->nick); /* Maintain exact case of nickname if the arg
                                  came from somewhere but the client's nick */

    clp = find_chan_link(cp, chan);
    if (clp == NULL) {
        if (cli != NULL)
            sendto_one(cli, RPL_FMT(cli, ERR_USERNOTINCHANNEL), cp->nick,
                    chan->name);
        return CHANMODE_FAIL;
    }

    switch (set) {
        case CHANMODE_SET:
            clp->flags |= ircd.cmodes.modes[mode].umask;
            return CHANMODE_OK;
        case CHANMODE_UNSET:
            clp->flags &= ~ircd.cmodes.modes[mode].umask;
            return CHANMODE_OK;
    }
    
    return CHANMODE_FAIL;
}

CHANMODE_QUERY_FUNC(chanmode_uflag_query) {
    *argused = 0;

    return CHANMODE_FAIL; /* you can't query these.. */
}

/*******************************************************************************
 * RFC1459 mode functions here
 ******************************************************************************/

/* Function to count how many bans in a channel a user matches against.  
 * We try to do the least number of comparisons possible (always 2,
 * maybe 3). */

static int check_bans(const struct channel_ban_list *list, const char *nick,
        const char *user, const char *host, const char *ip,
        const char *orighost) {
    struct channel_ban *cbp;
    bool check_orighost = false;
    int bans = 0;

    if (orighost != NULL && *orighost != '\0' && strcasecmp(host, orighost))
        check_orighost = true;

    /* Walk the list only once, checking each ban against the various
     * pieces of data provided.  If the nick or username don't match we skip
     * right away, then do two (or three) host checks.  First we check the
     * "display" host, then we do ip matching on the ip, and then we check
     * the orighost if required. */

    LIST_FOREACH(cbp, list, lp) {
        if (!match(cbp->nick, nick) || !match(cbp->user, user))
            continue; /* not a match */
        if (match(cbp->host, host) || ipmatch(cbp->host, ip)) {
            bans++;
            continue;
        } else if (check_orighost && match(cbp->host, orighost)) {
            bans++;
            continue;
        }
    }

    return bans;
}

/* bans.  rather complicated! */
static CHANMODE_FUNC(chanmode_ban) {
    struct channel_ban_list *banlist =
        (struct channel_ban_list *)chanmode_getdata(chan, mode);
    struct channel_ban *cbp;
    struct chanlink *clp;
    char nick[NICKLEN + 1];
    char user[USERLEN + 1];
    char host[HOSTLEN + 1];

    int cnt = 0;

    /* if we're not just clearing the data, check arg, and make sure they know
     * the argument is being used. */
    if (set != CHANMODE_CLEAR) {
        char *bang, *at;
        size_t nicklen, userlen, hostlen;

        if (arg == NULL) {
            /* if this is a client, send them the list of bans. */
            if (cli != NULL && MYCLIENT(cli)) {
                LIST_FOREACH(cbp, banlist, lp) {
                    sendto_one(cli, RPL_FMT(cli, RPL_BANLIST), chan->name,
                            cbp->nick, cbp->user, cbp->host, cbp->who,
                            cbp->when);
                }
                sendto_one(cli, RPL_FMT(cli, RPL_ENDOFBANLIST), chan->name);
            }
            return CHANMODE_NOP;
        }

        *argused = 1;
        /* if they're not ops, deny deny deny. */
        if (cli != NULL && !CHANOP(cli, chan) && !CLIENT_MASTER(cli))
            return ERR_CHANOPRIVSNEEDED;

        bang = strchr(arg, '!');
        at = strchr(arg, '@');
        /* here's the scoop: the bang can't be the first character of the
         * mask.  the at can't be the last character, and there has to be
         * something between the bang and the at. */
        if (bang == NULL || at == NULL || bang == arg ||
                *(at + 1) == '\0' || (bang + 1) == at || at < bang)
            return CHANMODE_FAIL; /* bogus mask */

        nicklen = bang - arg; /* length of text from beginning of the string
                                 to the '!' symbol. */
        userlen = at - (bang + 1); /* length from after the '!' to the '@'
                                    symbol */
        hostlen = strlen(at + 1); /* length from after the '@' to the end */

        /* Make sure nothing is too long, too */
        if (nicklen > NICKLEN || userlen > USERLEN || hostlen > HOSTLEN)
            return CHANMODE_FAIL; /* arguments are too long */

        strlcpy(nick, arg, nicklen + 1);
        strlcpy(user, bang + 1, userlen + 1);
        strlcpy(host, at + 1, hostlen + 1);
    }

    switch (set) {
        case CHANMODE_SET:
            LIST_FOREACH(cbp, banlist, lp) {
                if (++cnt > MAX_BANS_PER_CHANNEL && cli != NULL &&
                        MYCLIENT(cli)) {
                    sendto_one(cli, RPL_FMT(cli, ERR_BANLISTFULL),
                            chan->name, arg);
                    return CHANMODE_FAIL; /* no siree bob */
                }
                /* Previously I had tried to match the ban being set against
                 * other bans, but I'm not sure this behavior is desired so
                 * I'm taking it out for now... */
#if 0
                /* we only want to know if current bans cover this ban, not
                 * if this ban covers current ones, since it may cover
                 * other things too. */
                if (match(cbp->nick, nick) && match(cbp->user, user) &&
                        match(cbp->host, host))
                    return CHANMODE_FAIL; /* already set. */
#else
                if (!strcasecmp(cbp->nick, nick) &&
                        !strcasecmp(cbp->user, user) &&
                        !strcasecmp(cbp->host, host))
                    return CHANMODE_FAIL;
#endif
            }
            
            cbp = calloc(1, sizeof(struct channel_ban));
            strlcpy(cbp->nick, nick, NICKLEN + 1);
            strlcpy(cbp->user, user, USERLEN + 1);
            strlcpy(cbp->host, host, HOSTLEN + 1);
            if (cli != NULL)
                sprintf(cbp->who, "%s!%s@%s", cli->nick, cli->user, cli->host);
            else
                strcpy(cbp->who, ircd.me->name);
            cbp->when = me.now;
            cbp->type = CHANNEL_BAN_BAN;

            LIST_INSERT_HEAD(banlist, cbp, lp);

            /* count bans against all users in the channel. */
            LIST_FOREACH(clp, &chan->users, lpchan) {
                clp->bans = check_bans(banlist, clp->cli->nick,
                        clp->cli->user, clp->cli->host, clp->cli->ip,
                        clp->cli->orighost);
            }
            
            break;
        case CHANMODE_UNSET:
            LIST_FOREACH(cbp, banlist, lp) {
                if (!strcasecmp(cbp->nick, nick) &&
                        !strcasecmp(cbp->user, user) &&
                        !strcasecmp(cbp->host, host)) {
                    /* a winner. */
                    LIST_REMOVE(cbp, lp);
                    free(cbp);

                    /* count bans against all users in the channel. */
                    LIST_FOREACH(clp, &chan->users, lpchan) {
                        clp->bans = check_bans(banlist, clp->cli->nick,
                                clp->cli->user, clp->cli->host, clp->cli->ip,
                                clp->cli->orighost);
                    }
                    break;
                }
            }
            break;
        case CHANMODE_CLEAR:
            while ((cbp = LIST_FIRST(banlist)) != NULL) {
                LIST_REMOVE(cbp, lp);
                free(cbp);
            }
            break;
    }

    return CHANMODE_OK; /* a-okay. */
}

struct ban_query_state {
    bool started;
    struct channel_ban *curban;
};

CHANMODE_QUERY_FUNC(chanmode_ban_query) {
    struct channel_ban_list *banlist =
        (struct channel_ban_list *)chanmode_getdata(chan, mode);
    struct channel_ban *cbp;
    struct ban_query_state *bqs;
    char mask[BAN_MASK_LEN];
    size_t len;

    if (*state == NULL)
        *state = calloc(1, sizeof(struct ban_query_state));
    bqs = *state;

    if (bqs->started == false) {
        bqs->curban = LIST_FIRST(banlist);
        bqs->started = true;
    }
    
    /* always run this check because we might be done when we start.. */
    if (bqs->curban == NULL) {
        /* we're done here.. free bqs and let them leave the loop */
        free(bqs);
        *argused = 0;
        return CHANMODE_FAIL; /* all done! */
    }

    /* okay, we know we've either started or we finished and went home.. */
    cbp = bqs->curban;

    len = sprintf(mask, "%s!%s@%s", cbp->nick, cbp->user, cbp->host);
    if (len > (size_t)*argused) {
        *argused = -(len - *argused);
        /* do not advance the list as the consumer is expected not to use
         * this mode since they have no room for it. */
    } else {
        strcpy(arg, mask);
        *argused = len;
        /* advance the list since they promised us enough space! */
        bqs->curban = LIST_NEXT(cbp, lp);
    }

    return CHANMODE_OK;
}

static HOOK_FUNCTION(can_join_mode_b) {
    struct channel_check_args *ccap = (struct channel_check_args *)data;
    struct channel_ban_list *banlist =
        (struct channel_ban_list *)chanmode_getdata(ccap->chan,
                                                    ircd.cmodes.mode_ban);
    void *ret = (void *)HOOK_COND_OK; /* deny by default. */

    ccap->clp->bans = check_bans(banlist, ccap->cli->nick, ccap->cli->user,
            ccap->cli->host, ccap->cli->ip, ccap->cli->orighost);

    if (ccap->clp->bans)
        return (void *)ERR_BANNEDFROMCHAN;

    return ret; /* and return.. */
}

static HOOK_FUNCTION(can_send_mode_b) {
    struct channel_check_args *ccap = (struct channel_check_args *)data;

    if (ccap->clp == NULL)
        return (void *)HOOK_COND_NEUTRAL; /* doesn't effect us */
    if (ccap->clp->bans)
        return (void *)ERR_CANNOTSENDTOCHAN; /* banned, cannot send */

    return (void *)HOOK_COND_NEUTRAL;
}

static HOOK_FUNCTION(can_nick_mode_b) {
    struct channel_check_args *ccap = (struct channel_check_args *)data;
    struct channel_ban_list *banlist =
        (struct channel_ban_list *)chanmode_getdata(ccap->chan,
                                                    ircd.cmodes.mode_ban);

    if (ccap->clp == NULL) {
        log_debug("can_nick_mode_b() called when client wasn't in channel!");
        return (void *)HOOK_COND_NEUTRAL; /* not interested.. */
    }
    if (ccap->clp->bans)
       return (void *)ERR_BANNICKCHANGE; /* hope they handle this right. :) */

    /* if they're not banned, make sure the nickname change wouldn't result in
     * a ban either. */
    if (check_bans(banlist, ccap->extra, ccap->cli->user, ccap->cli->host,
                ccap->cli->ip, ccap->cli->orighost))
        return (void *)ERR_BANONCHAN;

    return (void *)HOOK_COND_NEUTRAL; /* eh. */
}

static HOOK_FUNCTION(can_send_mode_m) {
    struct channel_check_args *ccap = (struct channel_check_args *)data;

    if (chanmode_isset(ccap->chan, ircd.cmodes.mode_mod))
        return (void *)ERR_CANNOTSENDTOCHAN; /* unless overriden, no. */
    return (void *)HOOK_COND_NEUTRAL;
}

static HOOK_FUNCTION(can_send_mode_n) {
    struct channel_check_args *ccap = (struct channel_check_args *)data;

    /* if we're +n and they're not in the channel, don't let them send. */
    if (chanmode_isset(ccap->chan, ircd.cmodes.mode_nextern) &&
            ccap->clp == NULL)
        return (void *)ERR_CANNOTSENDTOCHAN;
    return (void *)HOOK_COND_NEUTRAL;
}

static HOOK_FUNCTION(can_show_mode_s) {
    struct channel_check_args *ccap = (struct channel_check_args *)data;

    /* if they're in the channel, it's always okay.  if the channel is +s it's
     * not okay unless they've got the see-hidden-channels privilege, and then
     * it's okay if they're opered.  if the channel isn't +s it's always okay
     * too. */
    if (ccap->clp != NULL)
        return (void *)HOOK_COND_OK; /* s'not a problem. */
    else if (chanmode_isset(ccap->chan, ircd.cmodes.mode_secret)) {
        if (CLIENT_OPERATOR(ccap->cli) &&
                BPRIV(ccap->cli, ircd.privileges.priv_shc))
            return (void *)HOOK_COND_ALWAYSOK; /* okay, but sketchy. */
        else
            return (void *)ERR_NOTONCHANNEL;
    } else
        return (void *)HOOK_COND_OK; /* it's not +s, okay by us. */
}

static CHANMODE_FUNC(chanmode_key) {

    if (set != CHANMODE_CLEAR) {
        if (arg == NULL)
            return CHANMODE_NOARG;
        if (cli != NULL && !CHANOP(cli, chan) && !CLIENT_MASTER(cli))
            return ERR_CHANOPRIVSNEEDED;
        *argused = 1;
    }

    switch (set) {
        case CHANMODE_SET:
            if (strchr(arg, ' ') != NULL || strchr(arg, ',') != NULL)
                return CHANMODE_FAIL; /* no spaces or commas in keys */
            if (cli == NULL && chanmode_isset(chan, mode)) {
                /* this is coming from a server... we do something silly here.
                 * if we have a key currently, we compare the two keys and use
                 * the one which is lexocographically greater (pretty
                 * arbitrary, eh?) */
                if (strcmp(arg, chanmode_getstrdata(chan, mode)) <= 0)
                    strncpy(chanmode_getstrdata(chan, mode), arg, PASSWDLEN);
                return CHANMODE_OK;
            }

            strncpy(chanmode_getstrdata(chan, mode), arg, PASSWDLEN);
            chanmode_setflag(chan, mode);
            break;
        case CHANMODE_UNSET:
            memset(chanmode_getstrdata(chan, mode), 0, PASSWDLEN + 1);
            chanmode_unsetflag(chan, mode);
            break;
    }

    return CHANMODE_OK;
}
static CHANMODE_QUERY_FUNC(chanmode_key_query) {
    size_t len;

    if (*state == NULL)
        *state = (void *)0x1; /* set it so we don't return more than once */
    else
        return CHANMODE_FAIL; /* already queried */

    if (!chanmode_isset(chan, mode)) {
        *argused = 0;
        return CHANMODE_FAIL; /* not set */
    } else {
        if ((len = strlen(chanmode_getstrdata(chan, mode))) >
                (size_t)*argused) {
            *argused = -(len - *argused);
            *state = NULL; /* they will have to ask again.. */
        } else {
            strncpy(arg, chanmode_getstrdata(chan, mode), PASSWDLEN);
            *argused = len;
        }
    }
    return CHANMODE_OK; /* it's set */
}


/* this function is hooked by the can_join event.  just compare the arg (if
 * any) to the key, and see if it's okay. */
static HOOK_FUNCTION(can_join_mode_k) {
    struct channel_check_args *ccap = (struct channel_check_args *)data;
    char *key = chanmode_getstrdata(ccap->chan, ircd.cmodes.mode_key);
    char *next, *cur;

    /* ccap->extra should be a comma separated list of keys (or NULL).  we try
     * each one in succession. */
    if (*key != '\0') {
        if (ccap->extra == NULL || *ccap->extra == '\0')
            return (void *)ERR_BADCHANNELKEY;
        cur = ccap->extra;
        next = strchr(cur, ',');
        while (cur != NULL) {
            if (!strncasecmp(key, cur,
                        (next == NULL ? PASSWDLEN : next - cur)))
                return (void *)HOOK_COND_OK; /* a match */
            if (next != NULL && *(next + 1) != '\0') {
                cur = next + 1;
                next = strchr(cur, ',');
            } else
                break;
        }
        return (void *)ERR_BADCHANNELKEY; /* no match, or no key */
    }

    return (void *)HOOK_COND_OK; /* okay to join. */
}

static CHANMODE_FUNC(chanmode_limit) {
    uint32_t limit;
    uint32_t *chl = (uint32_t *)chanmode_getdata(chan, mode);

    if (cli != NULL && !CHANOP(cli, chan) && !CLIENT_MASTER(cli))
        return ERR_CHANOPRIVSNEEDED;
    switch (set) {
        case CHANMODE_SET:
            if (arg == NULL)
                return CHANMODE_NOARG;
            *argused = 1;
            limit = str_conv_int(arg, 0);

            /* if the limit is bogus don't bother. */
            if (limit == 0)
                return CHANMODE_FAIL; /* bad number */
            
            *chl = limit;
            chanmode_setflag(chan, mode);
            break;
        case CHANMODE_UNSET:
            *chl = 0;
            chanmode_unsetflag(chan, mode);
            *argused = 0;
    }
  
    return CHANMODE_OK;
}

static CHANMODE_QUERY_FUNC(chanmode_limit_query) {
    char buf[16];
    size_t len;
    uint32_t *chl = (uint32_t *)chanmode_getdata(chan, mode);

    if (*state == NULL)
        *state = (void *)0x1; /* set it so we don't return more than once */
    else
        return CHANMODE_FAIL; /* already queried */
    if (!chanmode_isset(chan, mode))
        return CHANMODE_FAIL; /* not set */
    len = sprintf(buf, "%u", *chl);
    if (len > (uint32_t)*argused) {
        *argused = -(len - *argused);
        *state = NULL; /* they must ask again */
        return CHANMODE_OK;
    }
    strcpy(arg, buf);
    *argused = len;
    return CHANMODE_OK; /* it's set */
}


static HOOK_FUNCTION(can_join_mode_l) {
    struct channel_check_args *ccap = (struct channel_check_args *)data;

    /* if a limit is set and the number of people in the channel is greater
     * than the limit, don't let them in. */
    if (chanmode_getintdata(ccap->chan, ircd.cmodes.mode_limit) &&
            ccap->chan->onchannel >
            chanmode_getintdata(ccap->chan, ircd.cmodes.mode_limit))
        return (void *)ERR_CHANNELISFULL;

    return (void *)HOOK_COND_OK;
}

static HOOK_FUNCTION(can_act_mode_ov) {
    struct channel_check_args *ccap = (struct channel_check_args *)data;

    if (ccap->clp == NULL)
        return (void *)HOOK_COND_NEUTRAL;
    if (CLINKOP(ccap->clp) || CLINKVOICE(ccap->clp))
        return (void *)HOOK_COND_ALWAYSOK;  /* if they're op'd or voiced,
                                               they can always change nicks
                                               or send. */
    return (void *)HOOK_COND_NEUTRAL; /* not interested. */
}

/* vi:set ts=8 sts=4 sw=4 tw=76 et: */
