/*
 * chanmode.h: channel mode structures and prototypes
 * 
 * Copyright 2002-2007 the Ithildin Project.
 * See the COPYING file for more information on licensing and use.
 * 
 * $Id$
 */

#ifndef IRCD_CHANMODE_H
#define IRCD_CHANMODE_H

/*
 * channel mode fun.  this is a lot like the usermode request/release system,
 * except that it's a bit more complicated.  There are effectively six (yes,
 * six) types of channel modes.  The first are types A-E, and work as follows:
 * Modes of type A add or remove an item from a list.  They always have
 * parameters.  Modes of type B change a setting.  They always have a
 * parameter.  Modes of type C change a setting, but only have a parameter when
 * being set.  Modes of type E (yes E) change a setting, but only have a
 * parameter when being *unset*, and modes of type D change a boolean setting
 * and never take a parameter.  Last but not least are 'PREFIX' modes which
 * change a flag on a user in the channel.  All modes should take handler
 * functions which will be called when the mode is set/unset.  The handler
 * functions should be:
 * int func(client *, channel *, unsigned char, int, char *, int *)
 * the first argument should be the client changing the mode (if any, there
 * won't always be one!) this is handy for passing errors or providing custom
 * behavior.
 * the next argument is the channel on which the mode is being set, the third
 * is the mode character given (this allows several modes to all coalesce into
 * one function if that behavior is desired), the fourth will be one of five
 * values:
 * - CHANMODE_SET to set the mode normally
 * - CHANMODE_UNSET to unset the * mode normally
 * - CHANMODE_CLEAR to clear any data held by the mode (free allocated data,
 *   etc.  called when a channel or mode is being destroyed/release)
 * the fifth argument is the next argument in the mode command. the final holds
 * a return value which specifies whether the argument was used or not.  the
 * int should be set to 1 if the argument is used, and 0 if it isn't. 
 *
 * return values:
 * CHANMODE_OK: Change/request was successfully made
 * CHANMODE_NOP: This was an n-op (nothing happened)
 * CHANMODE_FAIL: The change or request failed
 * CHANMODE_NONEX: The mode does not exist
 * CHANMODE_NOARG: The mode needed an argument and none was given
 * number > 0: Equivalent to CHANMODE_FAIL with the value of a numeric to
 *             return to the user.
 */

#define CHANMODE_FUNC(x)                                                    \
    int x(client_t *cli, channel_t *chan, unsigned char mode, int set,      \
            char *arg __UNUSED, int *argused __UNUSED)
typedef int (*chanmode_func)(client_t *, channel_t *, unsigned char,
        int, char *, int *);

/*
 * Channel mode queries:  These work a lot like the chanmode_func stuff but
 * do not expect a client.  They exist simply to return state to people who
 * want to know.  The important item is the fifth argument, a private state
 * variable.  When querying begins the pointer value pointed to be state
 * must be NULL.  The value may never be changed, and any memory allocations
 * occuring on behalf of the queryer will be freed on the final query when
 * no more data can be reset/queried.
 *
 * return values:
 * CHANMODE_OK: Change/request was successfully made
 * CHANMODE_FAIL: The change or request failed
 * CHANMODE_NONEX: The mode does not exist
 */
#define CHANMODE_QUERY_FUNC(x)                                              \
   int x(channel_t *chan, unsigned char mode, char *arg __UNUSED,           \
           int *argused __UNUSED, void **state)
typedef int (*chanmode_query_func)(channel_t *, unsigned char, char *,
        int *, void **);

#define CHANMODE_FL_A       0x01
#define CHANMODE_FL_B       0x02
#define CHANMODE_FL_C       0x04
#define CHANMODE_FL_D       0x08
#define CHANMODE_FL_E       0x10
#define CHANMODE_FL_PREFIX  0x20

#define CHANMODE_SET        1
#define CHANMODE_UNSET      0
#define CHANMODE_CLEAR     -1

#define CHANMODE_OK         0
#define CHANMODE_FAIL      -1
#define CHANMODE_NOP       -2
#define CHANMODE_NONEX     -3
#define CHANMODE_NOARG     -4

void chanmode_init(bool);
void chanmode_deinit(bool);
uint64_t chanmode_request(unsigned char, unsigned char *, int,
        chanmode_func, chanmode_query_func, size_t, void *);
void chanmode_release(unsigned char);
void chanmode_update_funcs(unsigned char, chanmode_func,
        chanmode_query_func);
/* ways to set and unset channel modes.  channel modes are typically set by
 * their mode character, but can also be set by prefix in the case of userflag
 * types. */
int chanmode_set(unsigned char, client_t *, channel_t *, char *, int *);
int chanmode_setprefix(unsigned char, channel_t *, char *, int *);

int chanmode_unset(unsigned char, client_t *, channel_t *, char *, int *);
int chanmode_unsetprefix(unsigned char, channel_t *, char *, int *);

int chanmode_query(unsigned char, channel_t *, char *, int *, void **);

#define chanmode_setflag(chan, themode)                                       \
(chan->modes |= ircd.cmodes.modes[themode].mask)
#define chanmode_unsetflag(chan, themode)                                     \
(chan->modes &= ~ircd.cmodes.modes[themode].mask)

/* various checks to see if channel modes are set in specific ways.  you can
 * check to see if a flag-type mode is set on a channel, or you can check to
 * see if a user has a certain prefix.  lastly, you can get the data for a
 * specific mode (data type) */
#define chanmode_isset(chan, themode) \
((ircd.cmodes.modes[themode].avail == 0 &&                                \
  ircd.cmodes.modes[themode].mask) ?                                      \
 (chan->modes & ircd.cmodes.modes[themode].mask) :                        \
 0)

int chanmode_isprefix(channel_t *, client_t *, unsigned char);

#define chanlink_ismode(clp, mode)                                        \
(clp->flags & ircd.cmodes.modes[mode].umask)

#define chanmode_getdata(chan, themode)                                   \
((ircd.cmodes.modes[themode].avail == 0 &&                                \
  ircd.cmodes.modes[themode].mdi != NULL) ?                               \
 (chan->mdext + ircd.cmodes.modes[themode].mdi->offset) :                 \
 NULL)

#define chanmode_getintdata(chan, themode)                                \
((ircd.cmodes.modes[themode].avail == 0 &&                                \
  ircd.cmodes.modes[themode].mdi != NULL) ?                               \
 *(((uint32_t *)(chan->mdext +                                           \
             ircd.cmodes.modes[themode].mdi->offset))) :                  \
 0xFFFFFFFF)

#define chanmode_getstrdata(chan, themode)                                \
((ircd.cmodes.modes[themode].avail == 0 &&                                \
  ircd.cmodes.modes[themode].mdi != NULL) ?                               \
 ((char *)chan->mdext) + ircd.cmodes.modes[themode].mdi->offset :         \
 NULL)

/* this function returns a \0 terminated list of a client's prefixes in a
 * channel, in a static character array. */
char *chanmode_getprefixes(channel_t *, client_t *);

/* These two map prefixes to channel modes and vice versa. */
#define chanmode_prefixtomode(c)                                          \
(ircd.cmodes.pfxmap[(unsigned char)c] != NULL ?                           \
 ircd.cmodes.pfxmap[(unsigned char)c]->mode : '\0')                       
 
#define chanmode_modetoprefix(c)                                          \
(ircd.cmodes.modes[(unsigned char)c].umask ?                              \
 ircd.cmodes.modes[(unsigned char)c].prefix : '\0')

const char **chanmode_getmodes(channel_t *);

/* grab the offset of a mode. */
#define modeoffset(mode) ircd.cmodes.modes[mode].mdi->offset

struct chanmode {
    unsigned char mode;         /* the mode character */
    char    avail;              /* whether it is available or not */

    uint64_t mask;              /* the mask for this mode */
    short   umask;              /* the mask for a chanlink setting, if this is a
                                   chanuser mode */
    unsigned char prefix;       /* if this is a user-flag, it has a prefix, this is
                                   the prefix (specified by the caller in extdata).
                                   no error checking is done on this value. */
    chanmode_func changefunc;  /* the symbol/function to change the mode */
    chanmode_query_func queryfunc; /* the symbol/function to query the mode */
    int            flags;       /* flags given for this mode */
    struct mdext_item *mdi;     /* the mdext_item which describes the channel mode.
                                   allocated automatically.  */
};

/* some functions for generic channel mode handling */
CHANMODE_FUNC(chanmode_flag);
CHANMODE_QUERY_FUNC(chanmode_flag_query);
CHANMODE_FUNC(chanmode_uflag);
CHANMODE_QUERY_FUNC(chanmode_uflag_query);

/* this structure is used to hold bans on a channel. */
#define MAX_BANS_PER_CHANNEL 100
#define BAN_MASK_LEN NICKLEN + USERLEN + HOSTLEN + 3
LIST_HEAD(channel_ban_list, channel_ban);
struct channel_ban {
    char nick[NICKLEN + 1]; /* the three components of ban, split up for easier
                               processing. */
    char user[USERLEN + 1];
    char host[HOSTLEN + 1];
    char who[BAN_MASK_LEN + 1];
    time_t  when;       /* when the ban was set */
#define CHANNEL_BAN_BAN 0x01
    unsigned char type; /* reserved for various uses */

    LIST_ENTRY(channel_ban) lp;
};

/* some macros... the link variety are a lot faster if you've already taken
 * the time to see if they're in the channel (and have saved the chanlink
 * entry). */
#define CHANOP(cli, chan) chanmode_isprefix(chan, cli, '@')
#define CLINKOP(clp) chanlink_ismode(clp, ircd.cmodes.mode_op)
#define CHANVOICE(cli, chan) chanmode_isprefix(chan, cli, '+')
#define CLINKVOICE(clp) chanlink_ismode(clp, ircd.cmodes.mode_voice)

#endif
/* vi:set ts=8 sts=4 sw=4 tw=76 et: */
