/*
 * channel.c: channel structure management code
 * 
 * Copyright 2002-2007 the Ithildin Project.
 * See the COPYING file for more information on licensing and use.
 * 
 * This file provides the basic functions necessary to create/destroy/manage
 * channel structures, and lists of people with channels.
 */

#include <ithildin/ithildin.h>

#include "ircd.h"

IDSTRING(rcsid, "$Id$");

channel_t *create_channel(char *name) {
    channel_t *chan = calloc(1, sizeof(channel_t));

    /* allocate space for mode data too */
    chan->mdext = mdext_alloc(ircd.mdext.channel);

    strncpy(chan->name, name, ircd.limits.chanlen);
    chan->created = me.now;

    LIST_INSERT_HEAD(ircd.lists.channels, chan, lp);
    hash_insert(ircd.hashes.channel, chan);

    ircd.stats.channels++;

    hook_event(ircd.events.channel_create, chan);

    return chan;
}

void add_to_channel(client_t *cli, channel_t *chan, bool hook) {
    struct chanlink *clp;

    /* insert user to channel list, and vice versa */
    clp = malloc(sizeof(struct chanlink));
    clp->cli = cli;
    clp->chan = chan;
    clp->flags = clp->bans = 0;
    LIST_INSERT_HEAD(&chan->users, clp, lpchan);
    LIST_INSERT_HEAD(&cli->chans, clp, lpcli);

    chan->onchannel++;

    if (hook)
        hook_event(ircd.events.channel_add, clp);
}

void del_from_channel(client_t *cli, channel_t *chan, bool hook) {
    struct chanlink *clp = clp = find_chan_link(cli, chan);

    if (clp != NULL) {
        /* hook the del call first */
        if (hook)
            hook_event(ircd.events.channel_del, clp);

        /* remove from channel/client lists */
        LIST_REMOVE(clp, lpchan);
        LIST_REMOVE(clp, lpcli);
        free(clp);
        chan->onchannel--;
    }

    if (chan->onchannel == 0)
        destroy_channel(chan);
}

void destroy_channel(channel_t *chan) {
    unsigned char *s;
    int dummy;

    hook_event(ircd.events.channel_destroy, chan);
    
    /* clear out all the data possibly left over from channel modes. */
    s = ircd.cmodes.avail;
    while (*s) {
        ircd.cmodes.modes[*s].changefunc(NULL, chan, *s, CHANMODE_CLEAR, NULL, &dummy);
        s++;
    }

    /* we only expect to be called on an empty channel! */
    hash_delete(ircd.hashes.channel, chan);
    LIST_REMOVE(chan, lp);
    ircd.stats.channels--;
    mdext_free(ircd.mdext.channel, chan->mdext);
    free(chan);
}

struct chanlink *find_chan_link(client_t *cli, channel_t *chan) {
    struct chanlink *clp;

    LIST_FOREACH(clp, &cli->chans, lpcli) {
        if (clp->chan == chan)
            return clp;
    }

    return NULL;
}

int channel_check_access(client_t *cli, channel_t *chan, char *ext,
        event_t *ep) {
    struct channel_check_args cca;

    cca.cli = cli;
    cca.chan = chan;
    cca.clp = find_chan_link(cli, chan);
    cca.extra = ext;

    return hook_cond_event(ep, &cca);
}

char **channel_mdext_iter(char **last) {
    static int started = 0;
    channel_t *cp = *(channel_t **)last;

    if (cp == NULL) {
        if (started) {
            started = 0;
            return NULL;
        }
        cp = LIST_FIRST(ircd.lists.channels);
        started = 1;
    }
    if (cp == NULL)
        return NULL;

    *(channel_t **)last = LIST_NEXT(cp, lp);
    return (char **)&cp->mdext;
}

int chancmp(char *one, char *two, size_t len __UNUSED) {
    return istrcmp(ircd.maps.channel, one, two);
}
/* vi:set ts=8 sts=4 sw=4 tw=76 et: */
