/*
 * channel.h: channel (mode) structures and prototypes
 * 
 * Copyright 2002-2007 the Ithildin Project.
 * See the COPYING file for more information on licensing and use.
 * 
 * $Id$
 */

#ifndef IRCD_CHANNEL_H
#define IRCD_CHANNEL_H

LIST_HEAD(chanusers, chanlink);
LIST_HEAD(userchans, chanlink);

/* this structure is used to glue users and channels together.  one structure
 * is allocated per user/chan relationship, and is saved in a linked list on
 * the user's side, and on the channel's side. */
struct chanlink {
    client_t  *cli;
    channel_t *chan;
    short   flags;
    short   bans; /* for users only, stores how many bans they have against
                     them */

    LIST_ENTRY(chanlink) lpcli;
    LIST_ENTRY(chanlink) lpchan;
};

char **channel_mdext_iter(char **);

struct channel {
    char    name[CHANLEN + 1];         /* our channel's name. */
    time_t  created;                   /* the timestamp as well as creation time */

    unsigned int onchannel;            /* number of people on channel */
    int     flags;
    struct chanusers users;            /* list of users in channel */
    uint64_t modes;                    /* the flag-modes for the channel. */
    char    *mdext;                    /* mdext data */

    LIST_ENTRY(channel) lp;
};

/* create a channel.  give it a name, initially the channel will be empty,
 * so you use...*/
channel_t *create_channel(char *);

/* add_to_channel.  adds the given client to the given channel, fixes up all
 * structures properly.  the third argument is set if the caller wishes the
 * channel_add event to be hooked.  this is *almost* always the case. */
void add_to_channel(client_t *, channel_t *, bool);

/* del_from_channel.  removes the user from the given channel, if the
 * channel becomes empty, it is destroyed with...  the third argument is the
 * same as above. */
void del_from_channel(client_t *, channel_t *, bool);

/* destroy_channel.  destroys a channel and returns all the good stuff to
 * memory */
void destroy_channel(channel_t *);

/* find a channel by name */
#define find_channel(name) hash_find(ircd.hashes.channel, name)

/* finds a channel/user's chanlink entity and returns it (or NULL if it doesn't
 * exist).  Also, onchannel() is a wraparound checker to give a boolean return
 * based on this info. */
struct chanlink *find_chan_link(client_t *, channel_t *);
#define onchannel(cli, chan)                                               \
    (find_chan_link(cli, chan) != NULL ? 1 : 0)

#define check_channame(name)                                               \
    (name && (*name == '#') &&                                             \
     istr_okay(ircd.maps.channel, name))

/* this structure is used by channel_check_access to pass to its various hook
 * functions when checking if a channel can have an action performed on it (or
 * something.. */
struct channel_check_args {
    channel_t *chan;            /* the channel */
    client_t *cli;            /* the client */
    struct chanlink *clp;   /* if they're in the channel, this is the link. */
    char    *extra;            /* extra data */
};

/* this function determines whether a client (cli) can perform some action
 * related to a channel (chan).  it can be used in various manners and with
 * various events.  it calls the hooks for the specified event, which should
 * return one of the five CHANNEL_ statuses above, or an error numeric to send
 * to the user (which should be of the format '%s :...' unless your function is
 * going to have some special handling). */

/* these three are synonyms for the definitions in event.h */
#define CHANNEL_CHECK_OVERRIDE        HOOK_COND_SPASS
#define CHANNEL_CHECK_OK        HOOK_COND_PASS
#define CHANNEL_CHECK_NO        HOOK_COND_FAIL
int channel_check_access(client_t *, channel_t *, char *, event_t *);

/* this lets you check to see if a user can enter a channel. */
#define can_can_join_channel(cli, chan, arg)                                \
channel_check_access(cli, chan, arg, ircd.events.can_join_channel)
/* this lets you check to see if a user can see a channel's details. */
#define can_can_see_channel(cli, chan)                                        \
channel_check_access(cli, chan, NULL, ircd.events.can_see_channel)

/* this is like the above two, except that you also pass the message the user
 * is attempting to send to the channel. */
#define can_can_send_channel(cli, chan, msg)                                \
channel_check_access(cli, chan, msg, ircd.events.can_send_channel)

/* this is like the above three, except for nick changes in the channel.
 * hopefully this is it. :) */
#define can_can_nick_channel(cli, chan, nick)                                \
channel_check_access(cli, chan, nick, ircd.events.can_nick_channel);

/* chancmp, like nickcmp in client.c */
int chancmp(char *, char *, size_t);

#endif
/* vi:set ts=8 sts=4 sw=4 tw=76 et: */
