/*
 * connection.c: connection handling routines
 * 
 * Copyright 2002-2007 the Ithildin Project.
 * See the COPYING file for more information on licensing and use.
 * 
 * This file contains routines for accepting/checking/otherwise handling raw
 * socket connections.  Once connections are accepted they are passed to the
 * protocol system for further work.  This file also handles sending of data
 * using the queue mechanism, and other trivial socket related functions.
 */

#include <ithildin/ithildin.h>

/* for non-blocking dns and ident */
#include "../dns/dns.h"
#include "../dns/lookup.h"
#include "../ident/ident.h"

#include "ircd.h"

IDSTRING(rcsid, "$Id$");

static void connection_init_done(connection_t *);

void destroy_connection(connection_t *c, char *reason) {
    char msg[512];

    /* destroy their client/server entries. */
#ifdef DEBUG
    if (c->cli != NULL && c->srv != NULL)
        log_warn("connection %p has non-null client *and* server entry", c);
#endif
    /* destroy_client/server will set the cli/srv entries to NULL, then call us
     * again to really desotry the connection.  this makes it safe to call
     * 'destroy_connection' even if you're not sure what kind of connection you
     * have. */
    if (c->cli != NULL) {
        destroy_client(c->cli, reason);
        return;
    } else if (c->srv != NULL) {
        destroy_server(c->srv, reason);
        return;
    }

    /* Try to send them an error message if we can.  We assume someone has
     * tried to flush their sendq or doesn't think we should, so if it's
     * empty we send an error, otherwise we clean it off and don't bother
     * trying to send since other sends have already failed (and we infer
     * that this means the socket is not writeable) */
    if (STAILQ_EMPTY(&c->sendq) && reason != NULL && *reason != '\0') {
        snprintf(msg, 512, "\r\nERROR :Closing Link: %s (%s)\r\n", 
                c->host, reason);
        socket_write(c->sock, msg, strlen(msg));
    } else {
        while (STAILQ_FIRST(&c->sendq) != NULL)
            sendq_pop(c);
    }

    if (c->pass != NULL)
        free(c->pass);

    if (c->buf != NULL)
        free(c->buf);

    destroy_socket(c->sock);
    del_from_class(c);
    LIST_REMOVE(c, lp);
    free(c);
}

/* this function closes all 'unknown' (connected/unregistered) connections,
 * handy in various places. */
int close_unknown_connections(char *reason) {
    connection_t *cp;
    int count = 0;

    while (!LIST_EMPTY(ircd.connections.connected)) {
        cp = LIST_FIRST(ircd.connections.connected);
        destroy_connection(cp, reason);
        count++;
    }
    while (!LIST_EMPTY(ircd.connections.unregistered)) {
        cp = LIST_FIRST(ircd.connections.unregistered);
        destroy_connection(cp, reason);
        count++;
    }

    return count;
}

/* this function attempts to flush the send queue of a connection.  it returns
 * 1 if the client still exists (is not sendq'd off), or 0 otherwise. */
int sendq_flush(connection_t *conn) {
    struct sendq_item *sip;
    int ret;

    /* if the connection is writeable and has a send queue, push things off to
     * the socket. */
    if (conn->flags & IRCD_CONNFL_WRITEABLE) {
        while ((sip = STAILQ_FIRST(&conn->sendq)) != NULL) {
            ret = socket_write(conn->sock, sip->block->msg + sip->offset,
                    sip->block->len - sip->offset);
            if (ret <= 0)
                break;
            else if ((size_t)ret != sip->block->len - sip->offset) {
                sip->offset += ret;
                break;
            } else {
                conn->stats.sent += sip->block->len;
                conn->stats.psent++;
                sendq_pop(conn);
            }
        }
    }
    if (conn->flags & IRCD_CONNFL_NOSENDQ && conn->sendq_items == 0)
        conn->flags &= ~IRCD_CONNFL_NOSENDQ;
    else if (!(conn->flags & IRCD_CONNFL_NOSENDQ) &&
            conn->sendq_items > conn->cls->sendq) {
        destroy_connection(conn, "SendQ Exceeded");
        return 0;
    }

    return 1;
}

HOOK_FUNCTION(ircd_listen_hook) {
    connection_t *c = NULL;
    isocket_t *sp = NULL;
    isocket_t *listener = (isocket_t *)data;

    while ((sp = socket_accept(listener)) != NULL) {
        /* create a new connection, and immediately begin host and ident
         * checks */
        c = calloc(1, sizeof(connection_t));
        c->sock = sp;
        c->signon = c->last = me.now;
        sp->udata = c;
        get_socket_address(isock_raddr(sp), c->host, HOSTLEN + 1, NULL);

        add_to_class(LIST_FIRST(ircd.lists.classes), c);
        LIST_INSERT_HEAD(ircd.connections.connected, c, lp);
        add_hook(sp->datahook, ircd_connection_datahook);

        /* Check first for SSL goodies */
#ifdef HAVE_OPENSSL
        if (SOCKET_SSL(listener)) {
            if (socket_ssl_enable(c->sock)) {
                if (socket_ssl_accept(c->sock)) {
                    if (SOCKET_SSL_HANDSHAKING(c->sock)) {
                        /* handshake isn't done... we have to defer the
                         * event calling */
                        c->flags |= IRCD_CONNFL_SSLINIT;
                        continue;
                    }
                }
            }
            /* otherwise some error occured */
            destroy_connection(c, "");
            continue;
        }
#endif

        connection_init_done(c);

    }

    return NULL;
}

/* Called when the initial work on a connection should be done. */
static void connection_init_done(connection_t *c) {
    void **returns;
    int x;

    /* check for preliminary connection access */
    returns = hook_event(ircd.events.connection_connected, c);
    x = 0;
    while (x < hook_num_returns) {
        if (returns[x] != NULL) {
            /* this is not an allowed connection, drop it with the given
             * error message. */
            destroy_connection(c, returns[x]);
            break;
        }
        x++;
    }
}

bool connection_set_unregistered(connection_t *c) {
    void **returns;
    int x = 0;

    returns = hook_event(ircd.events.connection_unregistered, c);
    while (x < hook_num_returns) {
        if (returns[x] != NULL) {
            /* give the user an uninteresting error message, then dump our
             * structure. */
            destroy_connection(c, (char *)returns[x]);
            return false;
        }
        x++;
    }
    LIST_REMOVE(c, lp);
    LIST_INSERT_HEAD(ircd.connections.unregistered, c, lp);

    /* now stick them in the default protocol */
    c->proto = ircd.default_proto;
    c->last = me.now;
    /* and monitor them */
    socket_monitor(c->sock, SOCKET_FL_READ|SOCKET_FL_WRITE);

    return true;
}

HOOK_FUNCTION(ircd_connection_datahook) {
    isocket_t *s = (isocket_t *)data;
    connection_t *c = (connection_t *)s->udata;
    char msg[512];
    int ret;

#ifdef HAVE_OPENSSL
    if (c->flags & IRCD_CONNFL_SSLINIT) {
        if (SOCKET_ERROR(s)) {
            destroy_connection(c, "");
            return NULL;
        }
        c->flags &= ~IRCD_CONNFL_SSLINIT;
        connection_init_done(c);
        return NULL;
    }
    /* XXX: removed debug message for hooks when the connection isn't
     * completed.  we probably don't need this anymore... */
#endif
            
    /*
     * Read first, then write, then check errors.  If the read succeeds
     * (does not return IRCD_CONNECTION_CLOSED) we should be ensured that
     * the connection is still valid.
     */

    if (SOCKET_READ(s)) {
        ret = (int)c->proto->input(ep, data);
        /* We do this in a while loop because we may end up doing more than
         * one proto change (unlikely but the support for this is nearly
         * free so...) */
        while (ret == IRCD_PROTOCOL_CHANGED) {
            /* A protocol change implies two things to us:
             * 1) We may need to resize the buffer for this protocol.  We
             * can only resize up, not down!  In the down case we do the
             * wrong thing and close the connection instead of looping a
             * parser to clear the buffer. (XXX: worth fixing?)
             * 2) If there is any data in the buffer we should call the new
             * protocol's input function, protocol changes may leave
             * lingering data which would otherwise generate socket events
             * if we acted on it, so.. act on it here. */

            if (c->bufsize < c->proto->bufsize) {
                /* We must resize the buffer... we can happily just use
                 * realloc here, since nothing should depend on the address
                 * of c->buf between input calls */
                c->buf = realloc(c->buf, c->proto->bufsize);
                c->bufsize = c->proto->bufsize;
            }
            if (c->buflen > 0)
                /* We pass NULL as the event parameter for the input
                 * function as a way of saying "this was forcefully called
                 * outisde the normal event mechanism".  This is to tell the
                 * protocol handler that it should try to flush its current
                 * buffer even if there is no data from the socket.
                 * XXX: this is kind of a hack... */
                ret = (int)c->proto->input(NULL, data);
            else
                ret = 0; /* kill the loop :) */
        }

        if (ret == IRCD_CONNECTION_CLOSED)
            return NULL; /* nothing to do... */
    }

    if (SOCKET_WRITE(s))
        /* Only flag that we are writeable.  The writer hook takes care of
         * doing buffered writes for us outside the construct of this loop.
         * I thiml, potentially falsely, that it may be beneficial to batch
         * writes and reads in this manner. */
        c->flags |= IRCD_CONNFL_WRITEABLE;

    if (SOCKET_ERROR(s)) {
        if (c->srv != NULL) {
            sprintf(msg, "Read error from %s.  Closing link (%s)",
                    c->srv->name, socket_strerror(s));
            sendto_flag(SFLAG("GNOTICE"), msg);
            sendto_serv_butone(c->srv, NULL, ircd.me, NULL, "GNOTICE", ":%s",
                    msg);
        }
        /* XXX: Maybe swap in destroy_server and destroy_connection after it
         * with a source set on destroy_server of the server we're
         * destroying (so we don't send it a message for no reason) */
        destroy_connection(c, (s->err != 0 ? strerror(s->err) :
                    "Client closed connection"));
        return NULL;
    }

    return NULL;
}

HOOK_FUNCTION(ircd_writer_hook) {
    connection_t *cp, *cp2;

    /* handle sendqs for each of our three connection stages, in order */
    cp = LIST_FIRST(ircd.connections.connected);
    while (cp != NULL) {
        cp2 = LIST_NEXT(cp, lp);
        sendq_flush(cp);
        cp = cp2;
    }
    cp = LIST_FIRST(ircd.connections.unregistered);
    while (cp != NULL) {
        cp2 = LIST_NEXT(cp, lp);
        sendq_flush(cp);
        cp = cp2;
    }
    cp = LIST_FIRST(ircd.connections.clients);
    while (cp != NULL) {
        cp2 = LIST_NEXT(cp, lp);
        sendq_flush(cp);
        cp = cp2;
    }

    /* servers too */
    cp = LIST_FIRST(ircd.connections.servers);
    while (cp != NULL) {
        cp2 = LIST_NEXT(cp, lp);
        sendq_flush(cp);
        cp = cp2;
    }

    return NULL;
}

/* vi:set ts=8 sts=4 sw=4 tw=76 et: */
