/*
 * protocol.c: rudimentary protocol determination routines
 * 
 * Copyright 2002-2007 the Ithildin Project.
 * See the COPYING file for more information on licensing and use.
 * 
 * This file contains simple routines to allow connecting hosts to select a
 * protocol for use on the server, or have it otherwise determined for them.
 */

#include <ithildin/ithildin.h>

#include "ircd.h"

IDSTRING(rcsid, "$Id$");

/* this function, defined below, and the one defined immediately below here are
 * the default packet parsers.  The default packets resemble rfc1459 strongly.
 * The actual packet parser is a bit pickier about what gets passed along,
 * however. */
static int protocol_packet_parse(connection_t *cp, char *end);

/* Return codes from protocol_packet_parse */
#define PROTOCOL_PARSE_PROTOSET 1
#define PROTOCOL_PARSE_1459STOP 2

#define DEFAULT_BUFFER_SIZE 512
HOOK_FUNCTION(protocol_default_input) {
    isocket_t *sp = (isocket_t *)data;
    connection_t *cp = (connection_t *)sp->udata;
    int ret = 0;
    char *s;

    if (cp->buf == NULL || cp->bufsize == 0) {
        /* initialize the connection */
        cp->buf = malloc(DEFAULT_BUFFER_SIZE);
        cp->buflen = 0;
        cp->bufsize = DEFAULT_BUFFER_SIZE;
        cp->buf[0] = '\0';
    }

    if (cp->buflen < cp->bufsize)
        ret = socket_read(sp, cp->buf + cp->buflen,
                cp->bufsize - cp->buflen);
    if (ret < 0)
        return (void *)0; /* the error should be picked up by the handler. */
    else if (ret == 0)
        return (void *)0; /* nothing to do */

    cp->buflen += ret;

    cp->stats.recv += ret;
    /* the rest of this was yanked almost verbatim from proto/rfc1459.c, I had
     * done it a different way, but it ate a flaming bag of acidli when
     * connections to servers were attempted.  Now I've split this into a
     * pretty straightforward rfc1459 packet parser, and a command handler
     * function which lives below */

    while (cp->buflen > 0) {
        /* if the buffer is full but no terminating character (\n) is found, we
         * cheat */
        if (cp->buflen == cp->bufsize) {
            cp->buf[cp->bufsize - 1] = '\n';
            s = cp->buf + cp->bufsize - 1;
        } else {
            cp->buf[cp->buflen] = '\0';
            s = strchr(cp->buf, '\n');
            if (s == NULL)
                return (void *)0; /* no command terminator found and we're out
                                     of data */
        }
        /* make a local copy of the packet, then pass it off to the parser.
         * Bump the rest of the contents of 'buf' down if we read more than we
         * have parsed. */
        /* null-terminate and get rid of the [\r]\n sequence */
        if (*(s - 1) == '\r')
            *(s - 1) = '\0';
        else
            *s = '\0';

        /* Parse the packet, we must stop parsing immediately if the
         * connection is closed.  Additionally, we stop parsing immediately
         * if we get the 1459 return because the code that returns it
         * modifies our buffer for us.  In that case we return proto changed
         * and let input be call for us above. */
        if ((ret = protocol_packet_parse(cp, s)) == IRCD_CONNECTION_CLOSED)
            return (void *)ret; /* stop parsing.. */
        else if (ret == PROTOCOL_PARSE_1459STOP)
            return (void *)IRCD_PROTOCOL_CHANGED;
        s++; /* increment s, now we see if our packet contains more data */
        if ((size_t)(s - cp->buf) < cp->buflen) {
            cp->buflen -= s - cp->buf;
            memcpy(cp->buf, s, cp->buflen);
        } else {
            cp->buflen = 0; /* all done */
            *cp->buf = '\0'; /* no leftovers! */
        }
        if (ret == PROTOCOL_PARSE_PROTOSET)
            return (void *)IRCD_PROTOCOL_CHANGED;
    }

    return (void *)0;
}

static int protocol_packet_parse(connection_t *cp, char *end) {
    char *s, emsg[128];
    int rfc1459 = 0; /* this is such a dumb hack */

    cp->stats.precv++; /* another packet received.. */
    s = cp->buf;
    if (*s == ':') /* prefixed command?  ignore it totally! */
        return 0;

    if (!strncasecmp(cp->buf, "QUIT", 4)) {
        destroy_connection(cp, "");
        return IRCD_CONNECTION_CLOSED;
    } else if (!strncasecmp(cp->buf, "PROTOCOL", 8)) {
        /* we received a 'protocol' command.  cool.  this makes our life easy,
         * just grab the stuff after and find a protocol if possible */
        s += 8;
        while (isspace(*s))
            s++;
        if (*s == '\0') {
            destroy_connection(cp, "bogus protocol command");
            return IRCD_CONNECTION_CLOSED;
        }

        /* otherwise, s points to their desired protocol */
    } else if (!strncasecmp(s, "NICK", 4) || !strncasecmp(s, "USER", 4) ||
                !strncasecmp(s, "PASS", 4)) {
            /* this is probably an rfc1459 connection */
        rfc1459 = 1;
        s = "rfc1459";
    } else {
        log_debug("received weird packet for unprotocoled connection: %s",
                cp->buf);
        return 0;
    }

    /* s now points to our chosen protocol! joy */
    cp->proto = find_protocol(s);
    if (cp->proto == NULL) {
        sprintf(emsg, "protocol %s is not supported", s);
        destroy_connection(cp, emsg);
        return IRCD_CONNECTION_CLOSED;
    }
        
    /* call the setup function for this connection */
    if (cp->proto->setup != NULL)
        cp->proto->setup(cp);
        
    if (rfc1459) {
        /* Re-adjust the buffer and return the special token to immediately
         * and successfully halt parsing (see above).  We do this because
         * the protocol itself needs to receive the message we hooked onto
         * for 1459 verification. */
        if (*(end - 1) == '\0')
            *(end - 1) = '\r';
        *end = '\n';

        return PROTOCOL_PARSE_1459STOP;
    }

    return PROTOCOL_PARSE_PROTOSET;
}

protocol_t *find_protocol(char *name) {
    protocol_t *pp;

    LIST_FOREACH(pp, ircd.lists.protocols, lp) {
        if (!strcasecmp(pp->name, name))
            return pp;
    }
    return NULL;
}

int add_protocol(char *name) {
    char mname[PATH_MAX];
    protocol_t *pp;

    sprintf(mname, "ircd.protocol.%s", name);
    if (find_module(mname) == NULL &&
            !load_module(mname, MODULE_FL_CREATE|MODULE_FL_QUIET)) {
        log_error("unable to load module for protocol %s", name);
        return 0;
    }
        
    pp = calloc(1, sizeof(protocol_t));
    pp->name = strdup(name);
    pp->dll = find_module(mname);
    LIST_INSERT_HEAD(ircd.lists.protocols, pp, lp);

    update_protocol(pp->name);

    if (pp->input == NULL || pp->output == NULL) {
        remove_protocol(pp->name);
        return 0;
    }

    return 1;
}

/* this function updates the pointers in a protocol (most useful after a
 * reloading of modules */
void update_protocol(char *name) {
    protocol_t *pp = find_protocol(name);
    uint64_t *i64p;

    /* if it it doesn't exist, try to add it */
    if (pp == NULL) {
        if (!add_protocol(name))
            return;
        else
            pp = find_protocol(name);
    }

    log_debug("updating protocol %s", pp->name);
    pp->input = (hook_function_t)module_symbol(pp->dll, "input");
    pp->output = (protocol_output_func)module_symbol(pp->dll, "output");
    pp->setup = (void (*)(connection_t *))module_symbol(pp->dll, "setup");

    /* not all protocols will have these (specifically, client protocols will
     * not have them) */
    pp->register_user = (void (*)(connection_t *, client_t *))
        module_symbol(pp->dll, "register_user");
    pp->sync_channel = (void (*)(connection_t *, channel_t *))
        module_symbol(pp->dll, "sync_channel");

    if (pp->register_user != NULL || pp->sync_channel != NULL) {
        if (pp->register_user == NULL || pp->sync_channel == NULL)
            log_error("missing protocol functions in module for "
                    "protocol %s", name);
    }

    if (pp->input == NULL || pp->output == NULL)
        log_error("missing input or output function in module for "
                "protocol %s", name);

    if ((i64p = (uint64_t *)module_symbol(pp->dll, "protocol_flags")) != NULL)
        pp->flags = *i64p;

    if ((i64p = (uint64_t *)module_symbol(pp->dll, "protocol_buffer_size")) != NULL)
        pp->bufsize = *i64p;
    else
        pp->bufsize = DEFAULT_BUFFER_SIZE;
}

/* this removes a protocol.  basically it closes all connections in that
 * protocol and then removes it from the list. */
void remove_protocol(char *name) {
    protocol_t *pp = find_protocol(name);
    connection_t *cp, *cp2;

    if (pp == NULL)
        return;

    if (pp->dll->flags & MODULE_FL_RELOADING)
        return; /* don't do this here.. */

    cp = LIST_FIRST(ircd.connections.connected);
    while (cp != NULL) {
        cp2 = LIST_NEXT(cp, lp);
        if (cp->proto == pp)
            destroy_connection(cp, "protocol removed");
        cp = cp2;
    }
    cp = LIST_FIRST(ircd.connections.unregistered);
    while (cp != NULL) {
        cp2 = LIST_NEXT(cp, lp);
        if (cp->proto == pp)
            destroy_connection(cp, "protocol removed");
        cp = cp2;
    }

    cp = LIST_FIRST(ircd.connections.clients);
    while (cp != NULL) {
        cp2 = LIST_NEXT(cp, lp);
        if (cp->proto == pp)
            destroy_client(cp->cli, "protocol removed");
        cp = cp2;
    }
    cp = LIST_FIRST(ircd.connections.servers);
    while (cp != NULL) {
        cp2 = LIST_NEXT(cp, lp);
        if (cp->proto == pp)
            destroy_server(cp->srv, "protocol removed");
        cp = cp2;
    }

    LIST_REMOVE(pp, lp);
    free(pp->name);
    free(pp);
}

/* vi:set ts=8 sts=4 sw=4 tw=76 et: */
