/*
 * usermode.h: client mode structure declarations
 * 
 * Copyright 2002-2007 the Ithildin Project.
 * See the COPYING file for more information on licensing and use.
 * 
 * $Id$
 */

#ifndef IRCD_USERMODE_H
#define IRCD_USERMODE_H

/* mode stuff here.  modules/systems should acquire and release modes as they
 * need them.  You may not get the mode you asked for, if something else is
 * already using it, so don't assume that you did.  The function will return
 * the bitmask used to check for the mode on a user's structure, and also place
 * the actual mode letter in a character, passed as a pointer.  If you don't
 * care what mode you get specify '\0' as the suggested field.  Also, if you
 * want all users who set your specific mode to be placed in a 'send flag'
 * group (see send.[ch]) pass the value of the flag to use, if not pass -1.
 * The last argument is the name of the function used to determine whether a
 * user can set/unset the node.  The arguments are the client setting the mode,
 * the client on which the mode is being set, the mode character, an integer
 * specifying whether the mode is being set or unset, a pointer to an argument
 * (may be NULL), and a pointer to allow the function to specify whether the
 * mode used the argument or not. */
#define USERMODE_FUNC(x)                                                   \
int x(client_t *by __UNUSED, client_t *cli, unsigned char mode __UNUSED,   \
        int set __UNUSED, char *arg __UNUSED, int *argused __UNUSED)
typedef int (*usermode_func)(client_t *, client_t *, unsigned char, int,
        char *, int *);

/* helper functions for usermode initialization/cleanup */
void usermode_init(bool);
void usermode_deinit(bool);

uint64_t usermode_request(unsigned char, unsigned char *, int, int,
        usermode_func);

/* use this function to release a mode if you no longer care about it being set
 * on users.  You are still responsible for clearing out the group if the mode
 * had one */
void usermode_release(unsigned char);

/* this function provides a way to update the assigned function for a
 * usermode */
void usermode_update_func(unsigned char, usermode_func);

/* this function returns a mode string from the passed 64bit integer. the
 * string is statically allocated inside the function, if global is unset, get
 * all usermodes, otherwise, get only global modes */
unsigned char *usermode_getstr(uint64_t, char);

/* this does the opposite of the above, it returns a mask from a given string
 * of modes, if global is non-false, the mask will only contain 'global' flags,
 * otherwise, all flags are returned.  */
uint64_t usermode_getmask(unsigned char *, char);

/* this performs a 'diff' on the modes, and returns a string reflecting the
 * change from old modes to new modes, and places the result in 'result' */
void usermode_diff(uint64_t, uint64_t, char *, char);

/* this does all the work to set a mode on a user.  it checks to see if they
 * can set it, and places them in any appropriate groups.  if the function
 * succeeds it returns 1, if the user cannot set the mode it returns 0 */
int usermode_set(unsigned char, client_t *, client_t *, char *, int *);

/* this does like above, but unsetting instead of setting. */
int usermode_unset(unsigned char, client_t *, client_t *, char *, int *);

/* returns positive if a given mode is valid and set on a client */
#define usermode_isset(client, themode)                                       \
    ((ircd.umodes.modes[themode].avail == 0) ?                                \
     (client->modes & ircd.umodes.modes[themode].mask) :                      \
     0)

/* we know we will create some modes, create special macros to check them */
#define CLIENT_INVISIBLE(client) (client->modes & ircd.umodes.modes['i'].mask)
#define CLIENT_OPERATOR(client) (client->modes & ircd.umodes.modes['o'].mask)

struct usermode {
    unsigned char mode;             /* the actual mode */
    char    avail;                  /* 1 if available, 0 otherwise */
#define USERMODE_FL_GLOBAL  0x1     /* the usermode is spread across the
                                       network */
#define USERMODE_FL_OPER    0x2     /* the usermode is operator only */
#define USERMODE_FL_PRESERVE 0x4    /* preserve the mode once set unless
                                       explicitly unset by the user */
    int     flags;                  /* flags for the usermode */
    uint64_t mask;                  /* the bitmask for the mode */
    usermode_func changefunc;       /* the changer function for the mode */
    int     sflag;                  /* send flag (if any) for this mode. */
};

#endif
/* vi:set ts=8 sts=4 sw=4 tw=76 et: */
