/*
 * superstring.h: Single point of inclusion for using superstrings
 * 
 * Copyright 2007 Chip Norkus
 * See the COPYING file for more information on licensing and use.
 * 
 * $Id$
 */

#ifndef SUPERSTRING_H
#define SUPERSTRING_H

/* Files to include for other consumers. */
#include <stdbool.h>
#include <stddef.h>
#include <stdint.h>
#include <stdlib.h>

/* ****************************************************************************
 * SuperString Global Macros/Defines.  These are various behavioral changes
 * to the library.
 * - SUPERSTRING_DEBUG: If defined will add additional space in structures
 * and memory to debug SuperString use.
 * ****************************************************************************/
/* #undef SUPERSTRING_DEBUG */

/* ****************************************************************************
 * SuperString magic numbers.  Used for debugging bogus writes.
 * ****************************************************************************/
#ifdef SUPERSTRING_DEBUG
# define SUPERSTRING_MAGIC1             (0xdeadf00dbeefb0a7LL)
# define SUPERSTRING_MAGIC2             (0x38517de853a905bcLL)
# define SUPERSTRING_MAGIC3             (0x7592dead0368c0deLL)
# define SUPERSTRING_MAGIC_FILLSTR      ("\0\0\0\0\0\0\0\0")
# define superstring_magic_init(str)                                          \
    (str)->magic1 = SUPERSTRING_MAGIC1;                                       \
    (str)->magic2 = SUPERSTRING_MAGIC2
#else
# define superstring_magic_init(str) ((void *)0)
#endif

/* ****************************************************************************
 * SuperString Global Flags.  These are set program-wide and used by all the
 * SuperString functions to dictate various behaviors.
 * ****************************************************************************/
/* Various debugging flags */
#define SUPERSTRING_DEBUG_MAGIC         (0x0001)

/* Asserts for various errors, instead of returning non-success. */
#define SUPERSTRING_ASSERT_NOWRITE      (0x0002 << 16)
#define SUPERSTRING_ASSERT_BADMAGIC     (0x0004 << 16)

/* ****************************************************************************
 * SuperString flags.  These flags dictate the type of the superstring, as
 * well as certain attributes about it.
 * ****************************************************************************/
/* Is the structure really a large superstring structure? */
#define SUPERSTRING_JUMBO               (0x0001)
/* A string marked read-only.  This flag is mutable */
#define SUPERSTRING_NOWRITE             (0x0002)
/* An immutable (permanently read-only) string */
#define SUPERSTRING_IMMUTABLE           (0x0004)
/* A fixed-size string which cannot grow once allocated. */
#define SUPERSTRING_FIXED               (0x0008)
/* A string which is heap-backed (created from allocated memory) */
#define SUPERSTRING_HEAP                (0x0010)

/* The last bit in each byte is reserved for the 'anonymous' flag.
 * Anonymous strings are like stack-allocated strings, but actually contain
 * the entirety of a (false) superstring in a sequence of characters stored
 * on the stack (check out the SS_TEMP macro).  We must reserve both bits so
 * that we can ignore the byte order of the host trivially and set the first
 * two characters (which comprise a uint16_t size) to the same thing. */
#define SUPERSTRING_ANONYMOUS_LOW       (0x0080)
#define SUPERSTRING_ANONYMOUS_HIGH      (0x8000)
#define SUPERSTRING_ANONYMOUS_STR       "\200\200"

/* Meta-flags which comprise multiple flags */
#define SUPERSTRING_ANONYMOUS           (SUPERSTRING_ANONYMOUS_LOW |          \
                                         SUPERSTRING_ANONYMOUS_HIGH)
#define SUPERSTRING_READONLY            (SUPERSTRING_NOWRITE    |             \
                                         SUPERSTRING_IMMUTABLE  |             \
                                         SUPERSTRING_ANONYMOUS)

/* ****************************************************************************
 * The two SuperString types.  superstring_jumbo_t is meant to be used for
 * extremely large strings and must be worked with explicitly.  Note that it
 * has a variable size.  On 64-bit systems (where size_t is that of
 * uint64_t) jumbo strings can be extremely large, whereas on 32-bit systems
 * size is limited to 4GB.
 * ****************************************************************************/
struct _superstring_jumbo_ {
    uint16_t flags;
#ifdef SUPERSTRING_DEBUG
    uint64_t magic1;
#endif
    size_t len;
    size_t size;
#ifdef SUPERSTRING_DEBUG
    uint64_t magic2;
#endif
};

struct _superstring_small_ {
    uint16_t flags;
#ifdef SUPERSTRING_DEBUG
    uint64_t magic1;
#endif
    uint16_t len;
    uint16_t size;
#ifdef SUPERSTRING_DEBUG
    uint64_t magic2;
#endif
};

typedef struct _superstring_small_ * superstring;

/* ****************************************************************************
 * SuperString declarators.  Various macros useful for declaring specialized
 * variants of SuperStrings.
 * - superstring_stack should be used for creating stack space backed
 *   SuperStrings.
 * - superstring_struct should be used for creating superstrings which have
 *   their buffers backed within their containing structure.  They must be
 *   initialized with superstring_struct_init.
 * - superstring_anonymous should be used to create anonymous superstrings,
 *   for convenience you can also use SS_TEMP to denote the same thing.
 * ****************************************************************************/
#define superstring_stack(var, bufsize)                                       \
    unsigned char _ ## var ## _SS_BUF[bufsize +                               \
                                      sizeof(struct _superstring_small_)];    \
    superstring var = (superstring)&_ ## var ## _SS_BUF;                      \
    superstring_magic_init(var);                                              \
    var->flags = SUPERSTRING_FIXED;                                           \
    var->len = 0;                                                             \
    var->size = bufsize

#define superstring_struct(var, bufsize)                                      \
    unsigned char _ ## var ## _SS_BUF[bufsize +                               \
                                      sizeof(struct _superstring_small_)];    \
    superstring var;

#define superstring_struct_init(st, var)                                      \
    (st)->var = (st)->_ ## var ## _SS_BUF;                                    \
    superstring_magic_init((st)->var);                                        \
    (st)->var->flags = SUPERSTRING_FIXED;                                     \
    (st)->var->len = 0;                                                       \
    (st)->size = sizeof((st)->var - sizeof(struct _superstring_small_))

#ifdef SUPERSTRING_DEBUG
/* fudgy for magic number values.  Nothing looks at these for anonymous
 * strings (see SUPERSTRING_MAGIC_FILLSTR). */
# define superstring_anonymous(str) (superstring)                             \
    (SUPERSTRING_ANONYMOUS_STR /* flags */                                    \
     SUPERSTRING_MAGIC_FILLSTR /* magic1 */                                   \
     "\0\0" /* length */                                                      \
     "\0\0" /* size */                                                        \
     SUPERSTRING_MAGIC_FILLSTR /* magic2 */                                   \
     str)
#else
# define superstring_anonymous(str) (superstring)                             \
    (SUPERSTRING_ANONYMOUS_STR /* flags */                                    \
     "\0\0" /* length */                                                      \
     "\0\0" /* size */                                                        \
     str)
#endif

#define SS_TEMP superstring_anonymous

/* ****************************************************************************
 * Error code/return values
 * ****************************************************************************/
enum _superstring_status_ {
    SS_ERROR_SUCCESS = 0,
    SS_ERROR_OUTOFSPACE,
    SS_ERROR_READONLY,
    SS_ERROR_INVALID_PARAMETER,
    SS_ERROR_HEAP_FAILURE,
    SS_ERROR_NOT_HEAP,
    SS_ERROR_NO_ANONYMOUS,

    /* leave as the last entry */
    SS_ERROR_UNKNOWN
};

extern int ss_errno;

/* ****************************************************************************
 * The various API calls available.  Some have been inlined because they are
 * relatively trivial.
 * Note that any calls ending in _cs are calls which take a C string
 * (unsigned char *) as one of their values.  These will wrap the string in
 * a temporary superstring value and then issue calls as one would expect.
 * C strings can never be modified, only used as read-only references for
 * convenience.
 * ****************************************************************************/
/* Get the C representation of a superstring value. */
static inline const unsigned char *ss_cstring(superstring str) {
    if (str == NULL) {
        ss_errno = SS_ERROR_INVALID_PARAMETER;
        return NULL;
    }

    ss_errno = SS_ERROR_SUCCESS;
    return (str->flags & SUPERSTRING_JUMBO ?
            ((const unsigned char *)str) + sizeof(struct _superstring_jumbo_) :
            ((const unsigned char *)str) + sizeof(struct _superstring_small_));
}

/* Retrieve the length or size of a string */
static inline size_t ss_length(superstring str) {
    if (str == NULL) {
        ss_errno = SS_ERROR_INVALID_PARAMETER;
        return 0; /* bad answer.. */
    }

    ss_errno = SS_ERROR_SUCCESS;
    return (size_t)(str->flags & SUPERSTRING_JUMBO ?
            ((struct _superstring_jumbo_ *)str)->len :
            str->len);
}

static inline size_t ss_size(superstring str) {
    if (str == NULL) {
        ss_errno = SS_ERROR_INVALID_PARAMETER;
        return 0; /* bad answer.. */
    }

    ss_errno = SS_ERROR_SUCCESS;
    return (size_t)(str->flags & SUPERSTRING_JUMBO ?
            ((struct _superstring_jumbo_ *)str)->size :
            str->size);
}

/* Retrieve or set the character at 'offset' in the string */
static inline unsigned char ss_char(superstring str, size_t index) {
    if (str == NULL || index >= str->len) {
        ss_errno = SS_ERROR_INVALID_PARAMETER;
        return 0;
    }
    return (str->flags & SUPERSTRING_JUMBO ?
            *((unsigned char *)str + sizeof(struct _superstring_jumbo_) + index) :
            *((unsigned char *)str + sizeof(struct _superstring_small_) + index));
}
static inline int ss_setchar(superstring str, size_t index, char c) {
    if (str == NULL || index >= str->len)
        return (ss_errno = SS_ERROR_INVALID_PARAMETER);

    if (str->flags & SUPERSTRING_JUMBO)
        *((unsigned char *)str + sizeof(struct _superstring_jumbo_) + index) = c;
    else
        *((unsigned char *)str + sizeof(struct _superstring_small_) + index) = c;

    return (ss_errno = SS_ERROR_SUCCESS);
}

/*****************************************************************************/
/* String creation / destruction.  The basic inline function can be used
 * for simple requests where a value is either a simple \0 terminated string
 * or NULL (no value for a blank string). */
superstring ss_create_ex(const unsigned char *value,
                         uint16_t flags,
                         size_t len,
                         size_t size);
static inline superstring ss_create(const unsigned char *value) {
    return ss_create_ex(value, 0, 0, 0);
}

int ss_destroy(superstring *str);

/*****************************************************************************/
/* Duplicate a superstring.  Will retain appropriate flags (those that are
 * not instance-bound such as anonymous or fixed-size */
static inline superstring ss_duplicate(superstring str) {
    return ss_create_ex(ss_cstring(str),
            str->flags & (SUPERSTRING_JUMBO | SUPERSTRING_NOWRITE),
            str->len, str->size);
}

/*****************************************************************************/
/* Shorten or lengthen a superstring by trimming or appending data.  For
 * trim, if the length is greater than the current length spaces (or a
 * specified character) are appended as necessary. */
int ss_trim_ex(superstring str, size_t length, char fill);
static inline int ss_trim(superstring str, size_t length) {
    return ss_trim_ex(str, length, ' ');
}

/*****************************************************************************/
/* Append data to a superstring.  There are several variants provided which
 * allow for setting maximum string length, using C strings, etc */
#define SS_APPEND_CSTRING           (0x80000000)
int ss_append_ex(superstring *lstr, superstring rstr, int flags, size_t length);

/* Append a string with no concern for length */
static inline int ss_append(superstring *lstr, superstring rstr) {
    return ss_append_ex(lstr, rstr, 0, 0);
}
static inline int ss_append_cs(superstring *lstr, const unsigned char *rstr) {
    return ss_append_ex(lstr, (superstring)rstr, SS_APPEND_CSTRING, 0);
}

/* Append a string but do not allow lstr to exceed 'length' */
static inline int ss_nappend(superstring *lstr, superstring rstr,
        size_t length) {
    return ss_append_ex(lstr, rstr, 0, length);
}
static inline int ss_nappend_cs(superstring *lstr, const unsigned char *rstr,
        size_t length) {
    return ss_append_ex(lstr, (superstring)rstr, SS_APPEND_CSTRING, length);
}

/*****************************************************************************/
/* Comparison function, with various wrappers:
   - ss_compare_ex: Compare with flags and length.  If length is 0 it will be
                    determined at compare time. */
#define SS_COMPARE_IGNORECASE       (0x0001)
#define SS_COMPARE_CSTRING          (0x80000000)
int ss_compare_ex(superstring lstr, superstring rstr,
        int flags, size_t length);

/* Simple wrapper, length is determined automatically */
static inline int ss_compare(superstring lstr, superstring rstr) {
    return ss_compare_ex(lstr, rstr, 0, 0);
}
static inline int ss_compare_cs(superstring lstr, const unsigned char *rstr) {
    return ss_compare_ex(lstr, (superstring)rstr, SS_COMPARE_CSTRING, 0);
}

/* Compare up to 'length' bytes between two strings */
static inline int ss_ncompare(superstring lstr, superstring rstr,
        size_t length) {
    return ss_compare_ex(lstr, rstr, 0, length);
}
static inline int ss_ncompare_cs(superstring lstr, const unsigned char *rstr,
        size_t length) {
    return ss_compare_ex(lstr, (superstring)rstr, SS_COMPARE_CSTRING, length);
}

/* Compare in a case-insensitive (for ASCII only!) fashion */
static inline int ss_icompare(superstring lstr, superstring rstr) {
    return ss_compare_ex(lstr, rstr, SS_COMPARE_IGNORECASE, 0);
}
static inline int ss_icompare_cs(superstring lstr, const unsigned char *rstr) {
    return ss_compare_ex(lstr, (superstring)rstr, SS_COMPARE_CSTRING, 0);
}

/* length limited insensitive comparison */
static inline int ss_nicompare(superstring lstr, superstring rstr,
        size_t length) {
    return ss_compare_ex(lstr, rstr, SS_COMPARE_IGNORECASE, length);
}
static inline int ss_nicompare_cs(superstring lstr, const unsigned char *rstr,
        size_t length) {
    return ss_compare_ex(lstr, (superstring)rstr,
            SS_COMPARE_CSTRING | SS_COMPARE_IGNORECASE, length);
}

/*****************************************************************************/
/* Functions to manipulate the read/write status of a string.  Note that
 * ss_readonly will tell you if a string cannot be changed because it
 * matches some read-only condition (immutable, anonymous, or nowrite).  It
 * does not tell you if a string can be set to writeable.  A call to
 * ss_setwriteable will fail if this is the case. */
static inline bool ss_readonly(superstring str) {
    if (str == NULL)
        return (ss_errno = SS_ERROR_INVALID_PARAMETER);
    return (str->flags & SUPERSTRING_READONLY);
}

static inline int ss_setreadonly(superstring str) {
    if (str == NULL)
        return (ss_errno = SS_ERROR_INVALID_PARAMETER);
    if ((str->flags & SUPERSTRING_READONLY) != SUPERSTRING_NOWRITE)
        return (ss_errno = SS_ERROR_READONLY);
    str->flags |= SUPERSTRING_NOWRITE;
    return (ss_errno = SS_ERROR_SUCCESS);
}

static inline int ss_setwriteable(superstring str) {
    if (str == NULL)
        return (ss_errno = SS_ERROR_INVALID_PARAMETER);
    if ((str->flags & SUPERSTRING_READONLY) != SUPERSTRING_NOWRITE)
        return (ss_errno = SS_ERROR_READONLY);
    str->flags &= ~SUPERSTRING_NOWRITE;
    return (ss_errno = SS_ERROR_SUCCESS);
}

/*****************************************************************************/
/* Function to cap a string at its current size.  Note that fixed cannot be
 * removed once applied for various reasons */
static inline int ss_setfixed(superstring str) {
    if (str == NULL)
        return (ss_errno = SS_ERROR_INVALID_PARAMETER);
    str->flags |= SUPERSTRING_FIXED;
    return (ss_errno = SS_ERROR_SUCCESS);
}

/*****************************************************************************/
/* Set allocation functions to use for superstrings.  By default malloc,
 * realloc, and free are used. */
typedef void *(*superstring_malloc_function)(size_t);
typedef void *(*superstring_realloc_function)(void *, size_t);
typedef void (*superstring_free_function)(void *);
int ss_allocator_set(superstring_malloc_function malloc_func,
                     superstring_realloc_function realloc_func,
                     superstring_free_function free_func);

#endif /* SUPERSTRING_H */
