/*
 * allocator.c: Allocation functions for SuperStrings
 * 
 * Copyright 2007 Chip Norkus.
 * See the COPYING file for more information on licensing and use.
 * 
 */

#include <superstring/superstring.h>
#include "superstring_private.h"

IDSTRING(rcsid, "$Id$");

/*
 * The growth of SuperStrings is pegged against their current size.  The
 * minimum growth is SUPERSTRING_GROW_MIN and the maximum if
 * SUPERSTRING_GROW_MAX.  Growth is tiered by a 4 bit shift of size (16x
 * multiplier).  Assuming the minimum growth is 32, strings larger than 512
 * bytes move to the next tier.  Each growth size is based on the previous
 * growth value shifted 2 bits (4x).  The table below provides a better
 * example/expanation:
 * size     growth
 * <512     32
 * <2048    128
 * <8192    512
 * >=8192   2048
 */
#define SUPERSTRING_GROW_MIN   32
#define SUPERSTRING_GROW_MAX 2048
static inline size_t ss_growbytes(size_t cursize) {
    size_t grow = SUPERSTRING_GROW_MIN;
    /* determine growth factor */
    while (cursize > grow << 4 && grow < SUPERSTRING_GROW_MAX)
        grow <<= 2;

    /* We tweak the return to give the overall size that 'cursize' should
     * morph to.  It will always be at least one byte larger than cursize,
     * and may be up to twice as large.  The growth boundaries are set
     * against multiple-of-two sizes in order to attempt to make good use of
     * allocators.  Most allocators do not attempt to hold odd sized data
     * tightly in memory like a jigsaw puzzle, instead opting for sizes
     * along the powers-of-two line up to the system's page size.  While
     * we don't know the page size we are likely to find it, and will
     * definitely meet a factor of it, using the growth calculator.  This
     * helps save space and maximize the use of the memory the allocator is
     * likely to give us. */
    return cursize + (grow - (cursize % grow));
}

/* This function will grow the data pointed to by an ss_internal structure
 * as need-be.  It will also correctly handle structure type transitioning
 * and assign the correct memory address to 'outstr' (if provided) */
int ss_grow(ss_internal *istr, superstring *outstr) {
    size_t grow;
    size_t memsize;
    void *oldmem;
    superstring str;
    
    assert(istr != NULL);

    /* only heap-backed strings may be grown.  additionally we must not grow
     * fixed-size strings */
    if ((istr->flags & SUPERSTRING_HEAP) == 0)
        ss_returnerror(SS_ERROR_NOT_HEAP);
    if (istr->flags & SUPERSTRING_FIXED)
        ss_returnerror(SS_ERROR_OUTOFSPACE);

    /* Do not munge additional size on a new string.  For strings we are
     * really growing, ss_growbytes will give us the next 'growth factor'
     * size that will satisfactorily consume the existing size and provide
     * extra space.  For new strings we do not honor the factor-of-two byte
     * boundary concept for two reasons:
     * 1) Fixed buffers should not have additional space
     * 2) Small values (common) may be smaller than the minimum allocator
     * size, while the minimum grow size may be larger.  In this case we may
     * save some memory. */
    if (istr->str != NULL)
        grow = ss_growbytes(istr->size);
    else
        grow = istr->size;

    /* ensure we move to jumbo if we need to */
    if (istr->size + grow > 0xffff)
        istr->flags |= SUPERSTRING_JUMBO;

    if (istr->flags & SUPERSTRING_JUMBO)
        memsize = sizeof(struct _superstring_jumbo_);
    else
        memsize = sizeof(struct _superstring_small_);
    memsize += grow;

    /* realloc will DTRT if istr->str is NULL and simply create new memory
     * for us. */
    oldmem = (void *)istr->str;
    if ((istr->str = realloc(oldmem, memsize)) == NULL) {
        free(oldmem);
        ss_returnerror(SS_ERROR_HEAP_FAILURE);
    }
    /* Initialize the data */
    if (oldmem == NULL)
        memset((void *)istr->str, 0, memsize);
    istr->size = grow;

    /* istr->str may have shifted to jumbo, if it has we need to convert the
     * memory as well.  This is a bit expensive unfortunately.   Note that
     * we must do this before ssi_to_ss lest data be trampled. */
    str = istr->str;
    if (istr->flags & SUPERSTRING_JUMBO &&
            !(str->flags & SUPERSTRING_JUMBO)) {
        /* We need to move the memory at the end of the structure forward.
         * We can safely call memmove here, and need only give it the actual
         * string length (which might save us a little time) */
        memmove((char *)str + sizeof(struct _superstring_jumbo_),
                (char *)str + sizeof(struct _superstring_small_),
                (size_t)istr->len);
    }

    /* map in the updated values.  Mostly we can do this by calling
     * ssi_to_ss, but we also need to update the data pointer of istr to
     * reflect the new address of the data. */
    ssi_to_ss(istr, istr->str);
    istr->data = ss_cstring(str);
    if (outstr != NULL)
        *outstr = istr->str;

    ss_returnerror(SS_ERROR_SUCCESS);
}

superstring ss_create_ex(const unsigned char *value,
                         uint16_t flags,
                         size_t len,
                         size_t size) {
    ss_internal istr;
    unsigned char *cstr;

    /* Massage arguments a bit */
    if (value == NULL)
        value = "";
    if (len == 0)
        len = strlen(value);

    istr.flags = flags | SUPERSTRING_HEAP;
    istr.len = len;
    istr.str = NULL;
    if (size < istr.len + 1) {
        /* Unless this is a fixed buffer we consider size to be an advisory
         * parameter.  If it's not right we don't whine. */
        if (flags & SUPERSTRING_FIXED) {
            ss_errno = SS_ERROR_INVALID_PARAMETER;
            return NULL;
        }
        
        size = istr.len + 1;
    }
    /* If the size is still zero give it a reasonable default */
    if (size == 0)
        size = SUPERSTRING_GROW_MIN;
    istr.size = size;

    if (ss_grow(&istr, NULL) != SS_ERROR_SUCCESS)
        return NULL;

    /* Copy our data in */
    cstr = (unsigned char *)ss_cstring(istr.str);
    memcpy((void *)cstr, value, istr.len);
    cstr[istr.len] = '\0';

    return istr.str;
}

int ss_destroy(superstring *str) {

    if (str == NULL || *str == NULL)
        ss_returnerror(SS_ERROR_INVALID_PARAMETER);

    if (((*str)->flags & SUPERSTRING_HEAP) == 0)
        ss_returnerror(SS_ERROR_NOT_HEAP);

    ss_free(*str);
    *str = NULL;

    ss_returnerror(SS_ERROR_SUCCESS);
}

/* vi:set ts=8 sts=4 sw=4 tw=76 et: */
